# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from typing import Callable, Dict, List, Literal, Optional, Tuple

from jax import numpy as jnp, tree as jtr
from jaxtyping import Array
from numpyro.distributions import Distribution

from ...parameters import Parameters as P
from ..constraints import any_constraint
from ..utils import (
    ExtendedSupportTransformedDistribution,
    JointDistribution,
    ScaledMixture,
)
from ._ncombination import (
    combine_distributions,
    create_beta_distributions,
    create_independent_spin_orientation_gaussian_isotropic,
    create_powerlaw_redshift,
    create_powerlaws,
    create_truncated_normal_distributions,
    create_uniform_distributions,
)


def _build_non_mass_distributions(
    N: int,
    component_type: Literal["pl", "g"],
    mass_distributions: List[Distribution],
    use_beta_spin_magnitude: bool,
    use_truncated_normal_spin_magnitude: bool,
    use_truncated_normal_spin_x: bool,
    use_truncated_normal_spin_y: bool,
    use_truncated_normal_spin_z: bool,
    use_tilt: bool,
    use_eccentricity: bool,
    use_mean_anomaly: bool,
    use_redshift: bool,
    use_cos_iota: bool,
    use_polarization_angle: bool,
    use_right_ascension: bool,
    use_sin_declination: bool,
    use_detection_time: bool,
    use_phi_1: bool,
    use_phi_2: bool,
    use_phi_12: bool,
    use_phi_orb: bool,
    params: Dict[str, Array],
    validate_args: Optional[bool] = None,
) -> List[Distribution]:
    """Build distributions for non-mass parameters.

    Parameters
    ----------
    N : int
        Number of components
    component_type : Literal[&quot;pl&quot;, &quot;g&quot;]
        type of component, either "pl" or "g"
    mass_distributions : List[Distribution]
        list of mass distributions
    use_spin : bool
        whether to include spin
    use_tilt : bool
        whether to include tilt
    use_eccentricity : bool
        whether to include eccentricity
    use_mean_anomaly : bool
        whether to include mean_anomaly
    use_redshift : bool
        whether to include redshift
    use_cos_iota : bool
        whether to include cos_iota
    use_polarization_angle : bool
        whether to include polarization_angle
    use_right_ascension : bool
        whether to include right_ascension
    use_sin_declination : bool
        whether to include sin_declination
    use_detection_time : bool
        whether to include detection_time
    use_phi_1 : bool
        whether to include phi_1
    use_phi_2 : bool
        whether to include phi_2
    use_phi_12 : bool
        whether to include phi_12
    use_phi_orb : bool
        whether to include phi_orb
    params : Dict[str, Array]
        dictionary of parameters
    validate_args : Optional[bool], optional
        whether to validate arguments, by default None

    Returns
    -------
    List[Distribution]
        list of distributions
    """
    build_distributions = mass_distributions
    # fmt: off
    _info_collection: List[Tuple[bool, str, Callable[..., List[Distribution]]]] = [
        (use_beta_spin_magnitude, P.PRIMARY_SPIN_MAGNITUDE.value, create_beta_distributions),
        (use_beta_spin_magnitude, P.SECONDARY_SPIN_MAGNITUDE.value, create_beta_distributions),
        (use_truncated_normal_spin_magnitude, P.PRIMARY_SPIN_MAGNITUDE.value, create_truncated_normal_distributions),
        (use_truncated_normal_spin_magnitude, P.SECONDARY_SPIN_MAGNITUDE.value, create_truncated_normal_distributions),
        (use_truncated_normal_spin_x, P.PRIMARY_SPIN_X.value, create_truncated_normal_distributions),
        (use_truncated_normal_spin_x, P.SECONDARY_SPIN_X.value, create_truncated_normal_distributions),
        (use_truncated_normal_spin_y, P.PRIMARY_SPIN_Y.value, create_truncated_normal_distributions),
        (use_truncated_normal_spin_y, P.SECONDARY_SPIN_Y.value, create_truncated_normal_distributions),
        (use_truncated_normal_spin_z, P.PRIMARY_SPIN_Z.value, create_truncated_normal_distributions),
        (use_truncated_normal_spin_z, P.SECONDARY_SPIN_Z.value, create_truncated_normal_distributions),
        # combined tilt distribution
        (use_tilt, P.COS_TILT_1.value + "_" + P.COS_TILT_2.value, create_independent_spin_orientation_gaussian_isotropic),
        (use_phi_1, P.PHI_1.value, create_uniform_distributions),
        (use_phi_2, P.PHI_2.value, create_uniform_distributions),
        (use_phi_12, P.PHI_12.value, create_uniform_distributions),
        (use_eccentricity, P.ECCENTRICITY.value, create_truncated_normal_distributions),
        (use_mean_anomaly, P.MEAN_ANOMALY.value, create_uniform_distributions),
        (use_redshift, P.REDSHIFT.value, create_powerlaw_redshift),
        (use_right_ascension, P.RIGHT_ASCENSION.value, create_uniform_distributions),
        (use_sin_declination, P.SIN_DECLINATION.value, create_uniform_distributions),
        (use_detection_time, P.DETECTION_TIME.value, create_uniform_distributions),
        (use_cos_iota, P.COS_IOTA.value, create_uniform_distributions),
        (use_polarization_angle, P.POLARIZATION_ANGLE.value, create_uniform_distributions),
        (use_phi_orb, P.PHI_ORB.value, create_uniform_distributions),
    ]
    # fmt: on

    # Iterate over the list of tuples and build distributions
    for use, param_name, build_func in _info_collection:
        if use:
            distributions = build_func(
                N=N,
                parameter_name=param_name,
                component_type=component_type,
                params=params,
                validate_args=validate_args,
            )
            build_distributions = combine_distributions(
                build_distributions, distributions
            )

    return build_distributions


def _build_pl_component_distributions(
    N: int,
    use_beta_spin_magnitude: bool,
    use_truncated_normal_spin_magnitude: bool,
    use_truncated_normal_spin_x: bool,
    use_truncated_normal_spin_y: bool,
    use_truncated_normal_spin_z: bool,
    use_tilt: bool,
    use_eccentricity: bool,
    use_mean_anomaly: bool,
    use_redshift: bool,
    use_cos_iota: bool,
    use_polarization_angle: bool,
    use_right_ascension: bool,
    use_sin_declination: bool,
    use_detection_time: bool,
    use_phi_1: bool,
    use_phi_2: bool,
    use_phi_12: bool,
    use_phi_orb: bool,
    params: Dict[str, Array],
    validate_args: Optional[bool] = None,
) -> List[JointDistribution]:
    """Build distributions for power-law components.

    Parameters
    ----------
    N : int
        Number of components
    use_spin : bool
        whether to include spin
    use_tilt : bool
        whether to include tilt
    use_eccentricity : bool
        whether to include eccentricity
    use_mean_anomaly : bool
        whether to include mean_anomaly
    use_redshift : bool
        whether to include redshift
    use_cos_iota : bool
        whether to include cos_iota
    use_polarization_angle : bool
        whether to include polarization_angle
    use_right_ascension : bool
        whether to include right_ascension
    use_sin_declination : bool
        whether to include sin_declination
    use_detection_time : bool
        whether to include detection_time
    use_phi_1 : bool
        whether to include phi_1
    use_phi_2 : bool
        whether to include phi_2
    use_phi_12 : bool
        whether to include phi_12
    use_phi_orb : bool
        whether to include phi_orb
    params : Dict[str, Array]
        dictionary of parameters
    validate_args : Optional[bool], optional
        whether to validate arguments, by default None

    Returns
    -------
    List[JointDistribution]
        list of JointDistribution
    """
    powerlaws = create_powerlaws(N=N, params=params, validate_args=validate_args)

    mass_distributions = jtr.map(
        lambda powerlaw: [powerlaw],
        powerlaws,
        is_leaf=lambda x: isinstance(x, ExtendedSupportTransformedDistribution),
    )

    build_distributions = _build_non_mass_distributions(
        N=N,
        component_type="pl",
        mass_distributions=mass_distributions,
        use_beta_spin_magnitude=use_beta_spin_magnitude,
        use_truncated_normal_spin_magnitude=use_truncated_normal_spin_magnitude,
        use_truncated_normal_spin_x=use_truncated_normal_spin_x,
        use_truncated_normal_spin_y=use_truncated_normal_spin_y,
        use_truncated_normal_spin_z=use_truncated_normal_spin_z,
        use_tilt=use_tilt,
        use_eccentricity=use_eccentricity,
        use_redshift=use_redshift,
        use_cos_iota=use_cos_iota,
        use_phi_12=use_phi_12,
        use_polarization_angle=use_polarization_angle,
        use_right_ascension=use_right_ascension,
        use_sin_declination=use_sin_declination,
        use_detection_time=use_detection_time,
        use_phi_1=use_phi_1,
        use_phi_2=use_phi_2,
        use_phi_orb=use_phi_orb,
        use_mean_anomaly=use_mean_anomaly,
        params=params,
        validate_args=validate_args,
    )

    return [
        JointDistribution(*dists, validate_args=validate_args)
        for dists in build_distributions
    ]


def _build_g_component_distributions(
    N: int,
    use_beta_spin_magnitude: bool,
    use_truncated_normal_spin_magnitude: bool,
    use_truncated_normal_spin_x: bool,
    use_truncated_normal_spin_y: bool,
    use_truncated_normal_spin_z: bool,
    use_tilt: bool,
    use_eccentricity: bool,
    use_mean_anomaly: bool,
    use_redshift: bool,
    use_cos_iota: bool,
    use_polarization_angle: bool,
    use_right_ascension: bool,
    use_sin_declination: bool,
    use_detection_time: bool,
    use_phi_1: bool,
    use_phi_2: bool,
    use_phi_12: bool,
    use_phi_orb: bool,
    params: Dict[str, Array],
    validate_args: Optional[bool] = None,
) -> List[JointDistribution]:
    """Build distributions for Gaussian components.

    Parameters
    ----------
    N : int
        Number of components
    use_spin : bool
        whether to include spin
    use_tilt : bool
        whether to include tilt
    use_eccentricity : bool
        whether to include eccentricity
    use_mean_anomaly : bool
        whether to include mean_anomaly
    use_redshift : bool
        whether to include redshift
    use_cos_iota : bool
        whether to include cos_iota
    use_polarization_angle : bool
        whether to include polarization_angle
    use_right_ascension : bool
        whether to include right_ascension
    use_sin_declination : bool
        whether to include sin_declination
    use_detection_time : bool
        whether to include detection_time
    use_phi_1 : bool
        whether to include phi_1
    use_phi_2 : bool
        whether to include phi_2
    use_phi_12 : bool
        whether to include phi_12
    use_phi_orb : bool
        whether to include phi_orb
    params : Dict[str, Array]
        dictionary of parameters
    validate_args : Optional[bool], optional
        whether to validate arguments, by default None

    Returns
    -------
    List[JointDistribution]
        list of JointDistribution
    """
    m1_dists = create_truncated_normal_distributions(
        N=N,
        parameter_name="m1",
        component_type="g",
        params=params,
        validate_args=validate_args,
    )
    m2_dists = create_truncated_normal_distributions(
        N=N,
        parameter_name="m2",
        component_type="g",
        params=params,
        validate_args=validate_args,
    )

    mass_distributions = jtr.map(
        lambda m1, m2: [m1, m2],
        m1_dists,
        m2_dists,
        is_leaf=lambda x: isinstance(x, Distribution),
    )

    build_distributions = _build_non_mass_distributions(
        N=N,
        component_type="g",
        mass_distributions=mass_distributions,
        use_beta_spin_magnitude=use_beta_spin_magnitude,
        use_truncated_normal_spin_magnitude=use_truncated_normal_spin_magnitude,
        use_truncated_normal_spin_x=use_truncated_normal_spin_x,
        use_truncated_normal_spin_y=use_truncated_normal_spin_y,
        use_truncated_normal_spin_z=use_truncated_normal_spin_z,
        use_tilt=use_tilt,
        use_eccentricity=use_eccentricity,
        use_redshift=use_redshift,
        use_cos_iota=use_cos_iota,
        use_phi_12=use_phi_12,
        use_polarization_angle=use_polarization_angle,
        use_right_ascension=use_right_ascension,
        use_sin_declination=use_sin_declination,
        use_detection_time=use_detection_time,
        use_phi_1=use_phi_1,
        use_phi_2=use_phi_2,
        use_phi_orb=use_phi_orb,
        use_mean_anomaly=use_mean_anomaly,
        params=params,
        validate_args=validate_args,
    )

    return [
        JointDistribution(*dists, validate_args=validate_args)
        for dists in build_distributions
    ]


def NPowerlawMGaussian(
    N_pl: int,
    N_g: int,
    use_beta_spin_magnitude: bool = False,
    use_truncated_normal_spin_magnitude: bool = False,
    use_truncated_normal_spin_x: bool = False,
    use_truncated_normal_spin_y: bool = False,
    use_truncated_normal_spin_z: bool = False,
    use_tilt: bool = False,
    use_eccentricity: bool = False,
    use_redshift: bool = False,
    use_cos_iota: bool = False,
    use_phi_12: bool = False,
    use_polarization_angle: bool = False,
    use_right_ascension: bool = False,
    use_sin_declination: bool = False,
    use_detection_time: bool = False,
    use_phi_1: bool = False,
    use_phi_2: bool = False,
    use_phi_orb: bool = False,
    use_mean_anomaly: bool = False,
    *,
    validate_args=None,
    **params,
) -> ScaledMixture:
    r"""Create a mixture of power-law and Gaussian components.

    This model has a lot of parameters, we can not list all of them here. Therefore,
    we are providing the general description of the each sub model and their parameters.

    .. important::

        Important information about this models are as follows:

        * The first :code:`N_pl` components are power-law and the next :code:`N_g` components are Gaussian.
        * Log rates are named as :code:`log_rate_{i}` where :code:`i` is the index of the component.
        * First :code:`N_pl` log rates are for power-law components and the next :code:`N_g` log rates are for Gaussian components.
        * All log rates are in terms of natural logarithm.


    .. note::

        **Mass distribution**: For powerlaw mass distribution is
        :class:`PowerlawPrimaryMassRatio` and for Gaussian we have
        :class:`~numpyro.distributions.truncated.TruncatedNormal` distribution.

        .. math:: (m_1, m_2) \sim \text{PowerlawPrimaryMassRatio}(\alpha, \beta, m_{\text{min}}, m_{\text{max}})

        .. math:: (m_1, m_2) \sim \mathcal{N}_{[a,b]}(\mu, \sigma^2)

        **Spin distribution**: Spin is taken from a beta distribution. The beta distribution
        is parameterized by :math:`\mu` and :math:`\sigma^2` where :math:`\mu` is the mean
        and :math:`\sigma^2` is the variance of the beta distribution.

        .. math:: \chi_i \sim \mathrm{Beta}(\mu, \sigma^2)

        .. warning::
            Not every choice of :math:`\mu` and :math:`\sigma^2` will result in a valid beta
            distribution. The beta distribution is only valid when :math:`\mu \in (0, 1)`,
            :math:`\sigma^2 \in (0, 0.25)`, and :math:`\mu(1-\mu) > \sigma^2`. Refer to the
            `link <https://stats.stackexchange.com/questions/12232/calculating-the-parameters-of-a-beta-distribution-using-the-mean-and-variance#comment926602_12239>`_
            for more information.

        **Tilt distribution**: Tilt is taken from a
        :class:`~numpyro.distributions.truncated.TruncatedNormal` distribution, with
        fixed mean :math:`\mu=1` and fixed bounds, :math:`a=-1` and :math:`b=1`.

        .. math:: \cos(\theta_i) \sim \mathcal{N}_{[-1, 1]}(\sigma^2\mid\mu=1)

        **Eccentricity distribution**: Eccentricity is taken from
        :class:`~numpyro.distributions.truncated.TruncatedNormal` distribution.

        .. math:: \varepsilon_i \sim \mathcal{N}_{[a,b]}(\mu, \sigma^2)


    .. attention::

        Interestingly, in :class:`~numpyro.distributions.truncated.TruncatedNormal`
        distribution, if any of the bounds are not provided, it will be set to
        :math:`\pm\infty`. For example, if we set :math:`\mu=0` and do not provide the
        upper bound, then the resulting distribution would be a
        `half normal <https://en.wikipedia.org/wiki/Half-normal_distribution>`_
        distribution.


    The naming of the parameters follows the following convention:

    .. code:: bash

        <parameter name>_<model parameter>_<component type>_<component number>

    with an exception for the powerlaw mass distribution where the
    :code:`<parameter name>` is ignored. For example, spin is taken from a beta
    distribution whose parameters are :code:`mean` and :code:`variance`. The naming
    convention for the spin parameters would be:

    .. code:: text

        chi[1-2]_mean_(pl|g)_[0-N_pl+N_g]
        chi[1-2]_variance_(pl|g)_[0-N_pl+N_g]

    Parameters
    ----------

    N_pl : int
        Number of power-law components
    N_g : int
        Number of Gaussian components
    use_spin : bool
        whether to include spin, defaults to False
    use_tilt : bool
        whether to include tilt, defaults to False
    use_eccentricity : bool
        whether to include eccentricity, defaults to False
    use_mean_anomaly : bool
        whether to include mean_anomaly, defaults to False
    use_redshift : bool
        whether to include redshift, defaults to False
    use_cos_iota : bool
        whether to include cos_iota, defaults to False
    use_polarization_angle : bool
        whether to include polarization_angle, defaults to False
    use_right_ascension : bool
        whether to include right_ascension, defaults to False
    use_sin_declination : bool
        whether to include sin_declination, defaults to False
    use_detection_time : bool
        whether to include detection_time, defaults to False
    use_phi_1 : bool
        whether to include phi_1, defaults to False
    use_phi_2 : bool
        whether to include phi_2, defaults to False
    use_phi_12 : bool
        whether to include phi_12, defaults to False
    use_phi_orb : bool
        whether to include phi_orb, defaults to False
    validate_args : Optional[bool], optional
        whether to validate arguments, defaults to None

    Returns
    -------
    ScaledMixture
        scaled mixture of distributions
    """
    if N_pl > 0:
        pl_component_dist = _build_pl_component_distributions(
            N=N_pl,
            use_beta_spin_magnitude=use_beta_spin_magnitude,
            use_truncated_normal_spin_magnitude=use_truncated_normal_spin_magnitude,
            use_truncated_normal_spin_x=use_truncated_normal_spin_x,
            use_truncated_normal_spin_y=use_truncated_normal_spin_y,
            use_truncated_normal_spin_z=use_truncated_normal_spin_z,
            use_tilt=use_tilt,
            use_eccentricity=use_eccentricity,
            use_redshift=use_redshift,
            use_cos_iota=use_cos_iota,
            use_phi_12=use_phi_12,
            use_polarization_angle=use_polarization_angle,
            use_right_ascension=use_right_ascension,
            use_sin_declination=use_sin_declination,
            use_detection_time=use_detection_time,
            use_phi_1=use_phi_1,
            use_phi_2=use_phi_2,
            use_phi_orb=use_phi_orb,
            use_mean_anomaly=use_mean_anomaly,
            params=params,
            validate_args=validate_args,
        )

    if N_g > 0:
        g_component_dist = _build_g_component_distributions(
            N=N_g,
            use_beta_spin_magnitude=use_beta_spin_magnitude,
            use_truncated_normal_spin_magnitude=use_truncated_normal_spin_magnitude,
            use_truncated_normal_spin_x=use_truncated_normal_spin_x,
            use_truncated_normal_spin_y=use_truncated_normal_spin_y,
            use_truncated_normal_spin_z=use_truncated_normal_spin_z,
            use_tilt=use_tilt,
            use_eccentricity=use_eccentricity,
            use_redshift=use_redshift,
            use_cos_iota=use_cos_iota,
            use_phi_12=use_phi_12,
            use_polarization_angle=use_polarization_angle,
            use_right_ascension=use_right_ascension,
            use_sin_declination=use_sin_declination,
            use_detection_time=use_detection_time,
            use_phi_1=use_phi_1,
            use_phi_2=use_phi_2,
            use_phi_orb=use_phi_orb,
            use_mean_anomaly=use_mean_anomaly,
            params=params,
            validate_args=validate_args,
        )

    if N_pl == 0 and N_g != 0:
        component_dists = g_component_dist
    elif N_g == 0 and N_pl != 0:
        component_dists = pl_component_dist
    else:
        component_dists = pl_component_dist + g_component_dist

    N = N_pl + N_g
    log_rates = jnp.stack([params[f"log_rate_{i}"] for i in range(N)], axis=-1)

    return ScaledMixture(
        log_rates,
        component_dists,
        support=any_constraint(
            [component_dists.support for component_dists in component_dists]
        ),
        validate_args=validate_args,
    )



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from typing import List, Optional, Tuple, Union

import jax
from jax import numpy as jnp
from jax.scipy.special import xlog1py
from jaxtyping import Array, ArrayLike, PRNGKeyArray
from numpyro import distributions as dist
from numpyro._typing import DistributionT
from numpyro.distributions import constraints, Delta, Distribution
from numpyro.distributions.util import validate_sample

from gwkokab.utils.tools import error_if
from kokab.utils.regex import match_all


class _DirichletElement(Distribution):
    """DirichletElement distribution for order < n_dimensions - 1 and n_dimensions > 1."""

    arg_constraints = {
        "order": constraints.nonnegative_integer,
        "n_dimensions": constraints.positive_integer,
    }
    pytree_data_fields = ("order", "n_dimensions", "sum_of_concentrations")

    def __init__(
        self,
        order: float,
        n_dimensions: float,
        sum_of_concentrations: ArrayLike,
        validate_args: Optional[bool] = None,
    ):
        self.order = order
        self.n_dimensions = n_dimensions
        self.sum_of_concentrations = sum_of_concentrations
        batch_shape = jnp.shape(jnp.array(self.sum_of_concentrations))
        super(_DirichletElement, self).__init__(
            batch_shape=batch_shape, validate_args=validate_args
        )

    @constraints.dependent_property(is_discrete=False, event_dim=0)
    def support(self):
        return constraints.interval(0.0, 1.0 - self.sum_of_concentrations)

    @validate_sample
    def log_prob(self, value: Array) -> Array:
        constant = self.n_dimensions - self.order - 1
        log_prob_unnorm = xlog1py(constant - 1, -self.sum_of_concentrations - value)  # type: ignore
        log_norm = xlog1py(constant, -self.sum_of_concentrations) - jnp.log(constant)  # type: ignore
        return log_prob_unnorm - log_norm

    def sample(self, key: PRNGKeyArray, sample_shape: Tuple[int, ...] = ()) -> Array:
        u = jax.random.uniform(key, shape=sample_shape)
        inv_constant = 1.0 / (self.n_dimensions - self.order - 1)
        return (1.0 - self.sum_of_concentrations) * (1.0 - jnp.power(u, inv_constant))  # type: ignore


def DirichletElement(
    order: int,
    n_dimensions: int,
    validate_args: Optional[bool] = None,
    **kwargs,
) -> DistributionT:
    r"""Conditional Dirichlet Distribution.

    A `DirichletElement` is a distribution over the :math:`N`-dimensional simplex, where
    :math:`N` is the :code:`n_dimensions` parameter. The :math:`n`-th order Dirichlet
    distribution given by the :code:`order`. The distribution is defined as,

    .. math::
        \forall n \in \{0, N-2\}, p(x_n \mid x_0, \cdots, x_{n-1}) =
        \frac{(N-n-1)(1-\alpha_n-x_n)^{N-n-1}}{(1-\alpha_n)^{N-n-1}},
        \qquad x_n \in [0, \alpha_n]

    where :math:`\alpha_n = \sum_{i=0}^{n-1} \alpha_i` and :math:`x_n` is the
    :code:`order`-th element of the simplex.

    .. math::
        p(x_{N-1}\mid x_0, \cdots, x_{N-2}) = \frac{1}{1-\alpha_{N-1}},
        \qquad x_{N-1} \in [0,\alpha_{N-1}]

    Parameters
    ----------
    order : int
        order of the DirichletElement
    n_dimensions : int
        number of dimensions
    validate_args : Optional[bool], optional
        whether to validate the arguments, by default None

    Returns
    -------
    DistributionT
        DirichletElement distribution

    Raises
    ------
    KeyError
        Missing concentration parameters for `DirichletElement` of order {order}
    """
    error_if(
        isinstance(order, int) is False or order < 0,
        msg="`order` must be a non-negative integer.",
    )
    error_if(
        isinstance(n_dimensions, int) is False or n_dimensions < 2,
        msg="`n_dimensions` must be an integer greater than 1. "
        "If your problem is 1D, use a Uniform prior instead.",
    )
    try:
        concentrations = [kwargs["concentration" + str(i)] for i in range(order)]
    except KeyError as e:
        raise KeyError(
            f"Missing concentration parameters for `DirichletElement` of order {order}"
        ) from e

    sum_of_concentrations = sum(concentrations, start=0.0)
    high = 1.0 - sum_of_concentrations

    if order == n_dimensions - 1:
        return Delta(v=high, log_density=0.0, validate_args=validate_args)

    return _DirichletElement(
        order=float(order),
        n_dimensions=float(n_dimensions),
        sum_of_concentrations=sum_of_concentrations,
        validate_args=validate_args,
    )


def _available_prior(name: str) -> DistributionT:
    """Get the available prior from numpyro distributions.

    Parameters
    ----------
    name : str
        name of the prior

    Returns
    -------
    DistributionT
        prior distribution

    Raises
    ------
    AttributeError
        If the prior is not found in numpyro distributions
    """
    gwkokab_priors = {"DirichletElement": DirichletElement}
    error_if(
        name not in dist.__all__ + list(gwkokab_priors.keys()),
        AttributeError,
        f"Prior {name} not found. Available priors are: "
        + ", ".join(dist.__all__ + list(gwkokab_priors.keys())),
    )
    if name in gwkokab_priors:
        return gwkokab_priors[name]
    return getattr(dist, name)


def _is_lazy_prior(prior_dict: dict[str, Union[str, float]]) -> bool:
    """Check if the prior is a lazy prior.

    Parameters
    ----------
    prior_dict : dict[str, Union[str, float]]
        dictionary of prior

    Returns
    -------
    bool
        True if the prior is a lazy prior, False otherwise
    """
    for value in prior_dict.values():
        if isinstance(value, str):
            return True
    return False


def get_processed_priors(params: List[str], priors: dict) -> dict:
    """Get the processed priors from a list of parameters. A processed prior is either
    an instantiated prior or a tuple of :code:`(jax.tree_util.Partial, lazy_vars)` where
    :code:`lazy_vars` is a dictionary of lazy variables.

    Parameters
    ----------
    params : List[str]
        list of parameters
    priors : dict
        dictionary of priors

    Returns
    -------
    dict
        dictionary of processed priors

    Raises
    ------
    ValueError
        if the prior value is invalid
    """
    matched_prior_params = match_all(params, priors)
    for param in params:
        error_if(param not in matched_prior_params, msg=f"Missing prior for {param}")

    for key, value in matched_prior_params.items():
        # if the value is not a dict, means its a constant/duplicate value
        if not isinstance(value, dict):
            continue
        value_cpy = value.copy()
        dist_type = value_cpy.pop("dist", None)
        error_if(
            not isinstance(dist_type, str) or dist_type == "",
            msg=f"Prior for '{key}' must specify a 'dist' string field.",
        )
        prior = _available_prior(dist_type)

        # if there are no lazy variables, instantiate the prior
        if not _is_lazy_prior(value_cpy):
            matched_prior_params[key] = prior(**value_cpy, validate_args=True)
            continue

        # separate the lazy variables and the non-lazy variables
        lazy_vars = {k: v for k, v in value_cpy.items() if isinstance(v, str)}
        non_lazy_vars = {k: v for k, v in value_cpy.items() if not isinstance(v, str)}
        matched_prior_params[key] = (
            jax.tree_util.Partial(prior, **non_lazy_vars, validate_args=True),
            lazy_vars,
        )
    return matched_prior_params



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from typing import Callable, Dict, Tuple, TypeAlias

import equinox as eqx
import jax
from jax import nn as jnn, numpy as jnp, random as jrd
from jax.scipy.linalg import cho_solve
from jaxtyping import Array, PRNGKeyArray
from numpyro._typing import DistributionT
from numpyro.distributions import Distribution
from numpyro.distributions.continuous import _batch_mahalanobis, tri_logabsdet
from numpyro.distributions.util import cholesky_of_inverse

from gwkokab.models.utils import JointDistribution, ScaledMixture


StateT: TypeAlias = Tuple[
    Array,  # old monte-carlo-estimate
    Array,  # old error square
    Array,  # old size
    PRNGKeyArray,  # old key
]
"""State of the Monte Carlo estimation process:
   (log_estimate, error_sq, sample_count, rng_key)."""


@jax.jit
def monte_carlo_log_estimate_and_error_sq(
    log_probs: Array, N: Array
) -> Tuple[Array, Array]:
    """Computes the Monte Carlo estimate and error for the given log probabilities.

    Parameters
    ----------
    log_probs : Array
        Log probabilities of the samples.
    N : Array
        Number of samples used for the estimate.

    Returns
    -------
    Tuple[Array, Array]
        Monte Carlo logarithm of estimate, and square of the Monte Carlo error.
    """
    mask = ~jnp.isneginf(log_probs)
    log_moment_1 = jnn.logsumexp(log_probs, where=mask, axis=-1) - jnp.log(N)
    moment_2 = jnp.exp(jnn.logsumexp(2.0 * log_probs, where=mask, axis=-1)) / N
    error_sq = (moment_2 - jnp.exp(2.0 * log_moment_1)) / (N - 1.0)
    return log_moment_1, error_sq


@jax.jit
def combine_monte_carlo_log_estimates(
    log_estimates_1: Array, log_estimates_2: Array, N_1: Array, N_2: Array
) -> Array:
    r"""Combine two Monte Carlo estimates into a single estimate using the formula:

    .. math::

        \hat{\mu} = \frac{N_1 \hat{\mu}_1 + N_2 \hat{\mu}_2}{N_1 + N_2}

    Parameters
    ----------
    estimates_1 : Array
        First Monte Carlo estimate :math:`\ln\hat{\mu}_1`.
    estimates_2 : Array
        Second Monte Carlo estimate :math:`\ln\hat{\mu}_2`.
    N_1 : Array
        Number of samples used for the first estimate :math:`N_1`.
    N_2 : Array
        Number of samples used for the second estimate :math:`N_2`.

    Returns
    -------
    Array
        Combined Monte Carlo estimate :math:`\ln\hat{\mu}`.
    """
    combined_log_estimate = jnp.logaddexp(
        jnp.log(N_1) + log_estimates_1, jnp.log(N_2) + log_estimates_2
    ) - jnp.log(N_1 + N_2)
    return combined_log_estimate


@jax.jit
def combine_monte_carlo_errors_sq(
    error_1_sq: Array,
    error_2_sq: Array,
    log_estimate_1: Array,
    log_estimate_2: Array,
    log_estimate_3: Array,
    N_1: Array,
    N_2: Array,
) -> Array:
    r"""Combine two Monte Carlo errors into a single error estimate using the formula:

    .. math::

        \hat{\epsilon}^2=\frac{1}{N_3(N_3-1)}\sum_{k=1}^{2}\left\{N_k(N_k-1)\hat{\epsilon}_k^2+N_k\hat{\mu}^2_k\right\}-\frac{1}{N_3-1}\hat{\mu}^2

    where, :math:`N_3 = N_1 + N_2` is the total number of samples.

    Parameters
    ----------
    error_1_sq : Array
        Square of error of the first Monte Carlo estimate :math:`\hat{\epsilon}_1^2`.
    error_2_sq : Array
        Square of error of the second Monte Carlo estimate :math:`\hat{\epsilon}_2^2`.
    log_estimate_1 : Array
        Estimate of the first Monte Carlo estimate :math:`\ln\hat{\mu}_1`.
    log_estimate_2 : Array
        Estimate of the second Monte Carlo estimate :math:`\ln\hat{\mu}_2`.
    log_estimate_3 : Array
        Estimate of the combined Monte Carlo estimate :math:`\ln\hat{\mu}`.
    N_1 : Array
        Number of samples used for the first estimate :math:`N_1`.
    N_2 : Array
        Number of samples used for the second estimate :math:`N_2`.

    Returns
    -------
    Array
        Combined Monte Carlo squared error estimate :math:`\hat{\epsilon}^2`.
    """
    N_3 = N_1 + N_2

    sum_prob_sq_1 = N_1 * ((N_1 - 1.0) * error_1_sq + jnp.exp(2.0 * log_estimate_1))
    sum_prob_sq_2 = N_2 * ((N_2 - 1.0) * error_2_sq + jnp.exp(2.0 * log_estimate_2))

    combined_error_sq = -jnp.exp(2.0 * log_estimate_3) / (N_3 - 1.0)
    combined_error_sq += (sum_prob_sq_1 + sum_prob_sq_2) / N_3 / (N_3 - 1.0)

    return combined_error_sq


@eqx.filter_jit
def mvn_samples(
    loc: Array, scale_tril: Array, n_samples: int, key: PRNGKeyArray
) -> Array:
    """Generate samples from a multivariate normal distribution using method from
    `numpyro.distributions.MultivariateNormal`.

    Parameters
    ----------
    loc : Array
        Mean vector of the multivariate normal distribution.
    cov : Array
        Covariance matrix of the multivariate normal distribution.
    n_samples : int
        Number of samples to generate.
    key : PRNGKeyArray
        JAX random key for sampling.

    Returns
    -------
    Array
        Samples drawn from the multivariate normal distribution.
    """
    eps = jrd.normal(key, shape=(n_samples, *loc.shape))
    samples = loc + jnp.squeeze(jnp.matmul(scale_tril, eps[..., jnp.newaxis]), axis=-1)
    return samples


@jax.jit
def mvn_log_prob_scaled(loc: Array, scale_tril: Array, value: Array) -> Array:
    # removing the constant term -0.5 * D * log(2pi), where D is the dimension,
    # because it is being cancelled out in the importance sampling weights
    M = _batch_mahalanobis(scale_tril, value - loc)
    return -0.5 * M - tri_logabsdet(scale_tril)


@jax.jit
def covariance_matrix(scale_tril: Array) -> Array:
    return jnp.matmul(scale_tril, jnp.swapaxes(scale_tril, -1, -2))


@jax.jit
def precision_matrix(scale_tril: Array) -> Array:
    identity = jnp.broadcast_to(jnp.eye(scale_tril.shape[-1]), scale_tril.shape)
    return cho_solve((scale_tril, True), identity)


@jax.jit
def is_positive_semi_definite(matrix: Array) -> Array:
    """Check if a matrix is positive semi-definite, by verifying that all its
    eigenvalues are non-negative.

    Parameters
    ----------
    matrix : Array
        The matrix of shape (..., N, N) to be checked.

    Returns
    -------
    Array
        Boolean array of shape (...) indicating whether each matrix is positive
        semi-definite.
    """
    return jnp.all(jnp.linalg.eigvals(matrix) >= 0, axis=-1)


def analytical_likelihood(
    dist_fn: Callable[..., Distribution],
    priors: JointDistribution,
    variables_index: Dict[str, int],
    poisson_mean_estimator: Callable[[ScaledMixture], Array],
    key: PRNGKeyArray,
    n_events: int,
    minimum_mc_error: float = 1e-2,
    n_samples: int = 100,
    n_checkpoints: int = 10,
    n_max_steps: int = 20,
) -> Callable[[Array, Dict[str, Array]], Array]:
    r"""Compute the analytical likelihood function for a model given its parameters.

    .. math::

        \mathcal{L}_{\mathrm{analytical}}(\Lambda,\kappa)\propto
        \exp\left(-\mu(\Lambda,\kappa)\right)
        \prod_{i=1}^{N}
        \iint\mathcal{G}(\theta,z|\boldsymbol{\mu}_i,\boldsymbol{\Sigma}_i)
        \rho(\theta,z\mid\Lambda,\kappa)d\theta dz

        \mathcal{L}_{\mathrm{analytical}}(\Lambda,\kappa)\propto
        \exp{\left(-\mu(\Lambda,\kappa)\right)}\prod_{i=1}^{N}\bigg<
        \rho(\theta_{i,j},z_{i,j}\mid\Lambda,\kappa)
        \bigg>_{\theta_{i,j},z_{i,j}\sim\mathcal{G}(\theta,z|\boldsymbol{\mu}_i,\boldsymbol{\Sigma}_i)}


    First we are optimizing the mean vector and covariance matrix of the multivariate
    normal distribution that approximates the product of event distributions (also a
    multivariate normal distribution) and phenomenological model distribution by
    Importance Sampling and Moment Matching. After that, we perform Variational
    Inference to optimize the mean vector of the multivariate normal distribution to
    minimize the Reverse KL divergence between the product of event distributions and
    the phenomenological model distribution.

    Parameters
    ----------
    dist_fn : Callable[..., Distribution]
        function that returns a Distribution instance
    priors : JointDistribution
        priors for the model parameters
    variables_index : Dict[str, int]
        mapping of variable names to their indices in the input array
    poisson_mean_estimator : Callable[[ScaledMixture], Array]
        function to compute the expected event rates
    n_events : int
        number of events
    key : PRNGKeyArray
        JAX random key for sampling
    minimum_mc_error : float, optional
        Minimum threshold for Monte Carlo error, by default 1e-2
    n_samples : int, optional
        Number of samples to draw from the multivariate normal distribution for each
        event to compute the likelihood, by default 100
    n_vi_steps: int, optional
        Number of steps for the variational inference optimization, by default 5
    learning_rate : float, optional
        Learning rate for the Adam optimizer used in the variational inference
        optimization, by default 1e-2
    n_checkpoints : int, optional
        Checkpoint interval for the checkpointed while-loop, by default 10. For more
        details, see
        [`equinox.internal.while_loop`](https://github.com/patrick-kidger/equinox/blob/main/equinox/internal/_loop/loop.py)
        and its [`"checkpointed"` variant](https://github.com/patrick-kidger/equinox/blob/main/equinox/internal/_loop/checkpointed.py).
    n_max_steps : int, optional
        Maximum number of steps for the optimization process, by default 20

    Returns
    -------
    Callable[[Array, Array], Array]
        A function that computes the log posterior probability of the model parameters
        given the data. The function takes two arguments: an array of model parameters
        and a second array (not used in this implementation).
    """
    minimum_mc_error_sq = minimum_mc_error**2

    def log_likelihood_fn(x: Array, data: Dict[str, Array]) -> Array:
        mean_stack = data["mean_stack"]
        scale_tril_stack = data["scale_tril_stack"]
        T_obs = data["T_obs"]

        mapped_params = {
            name: jax.lax.dynamic_index_in_dim(x, i, keepdims=False)
            for name, i in variables_index.items()
        }

        model_instance: DistributionT = dist_fn(**mapped_params)

        # μ = E_{Ω|Λ}[VT(ω)]
        expected_rates = poisson_mean_estimator(model_instance)

        rng_key = key

        hessian_log_prob = jax.jit(
            jax.vmap(jax.hessian(model_instance.log_prob), axis_size=n_events)
        )
        grad_log_prob = jax.jit(
            jax.vmap(jax.grad(model_instance.log_prob), axis_size=n_events)
        )

        fit_precision_matrix = precision_matrix(scale_tril_stack) - hessian_log_prob(
            mean_stack
        )
        fit_covariance_matrix = jnp.linalg.inv(fit_precision_matrix)

        is_psd = is_positive_semi_definite(fit_covariance_matrix)

        safe_fit_covariance_matrix = jnp.where(
            is_psd[..., jnp.newaxis, jnp.newaxis],
            fit_covariance_matrix,
            jnp.broadcast_to(
                jnp.eye(scale_tril_stack.shape[-1]),
                scale_tril_stack.shape,
            ),
        )

        fit_mean = mean_stack + jnp.where(
            is_psd[..., jnp.newaxis],
            jnp.squeeze(
                safe_fit_covariance_matrix
                @ jnp.expand_dims(grad_log_prob(mean_stack), axis=-1),
                axis=-1,
            ),
            jnp.zeros_like(mean_stack),
        )
        fit_scale_tril = jnp.where(
            is_psd[..., jnp.newaxis, jnp.newaxis],
            cholesky_of_inverse(fit_precision_matrix),
            scale_tril_stack,
        )

        mvn_log_prob_while_body = jax.vmap(
            mvn_log_prob_scaled, in_axes=(None, None, 0), axis_size=n_samples
        )

        def scan_fn(
            carry: Array, loop_data: Tuple[Array, Array, Array, Array, PRNGKeyArray]
        ) -> Tuple[Array, None]:
            mean, scale_tril, fit_mean_i, fit_scale_tril_i, rng_key_i = loop_data

            def while_body_fn(state: StateT) -> StateT:
                log_estimate_1, error_1_sq, N_1, rng_key = state
                N_2 = n_samples
                rng_key, subkey = jrd.split(rng_key)

                # data ~ G(θ, z | μ_f, Σ_f)
                data = mvn_samples(
                    loc=fit_mean_i,
                    scale_tril=fit_scale_tril_i,
                    n_samples=N_2,
                    key=subkey,
                )

                # log ρ(data | Λ, κ)
                model_instance_log_prob = jax.vmap(
                    model_instance.log_prob, axis_size=N_2
                )(data)

                # log G(θ, z | μ_i, Σ_i)
                event_mvn_log_prob = mvn_log_prob_while_body(mean, scale_tril, data)

                # log G(θ, z | μ_f, Σ_i)
                fit_mvn_log_prob = mvn_log_prob_while_body(
                    fit_mean_i, fit_scale_tril_i, data
                )

                log_prob = (
                    model_instance_log_prob + event_mvn_log_prob - fit_mvn_log_prob
                )
                log_estimate_2, error_2_sq = monte_carlo_log_estimate_and_error_sq(
                    log_prob,
                    N_2,  # type: ignore[arg-type]
                )
                log_estimate_3 = combine_monte_carlo_log_estimates(
                    log_estimate_1,
                    log_estimate_2,
                    N_1,
                    N_2,  # type: ignore[arg-type]
                )
                error_3_sq = combine_monte_carlo_errors_sq(
                    error_1_sq,
                    error_2_sq,
                    log_estimate_1,
                    log_estimate_2,
                    log_estimate_3,
                    N_1,
                    N_2,  # type: ignore[arg-type]
                )
                return log_estimate_3, error_3_sq, N_1 + N_2, rng_key

            state_0 = (
                -jnp.inf,  # starting log estimate is -inf
                jnp.zeros(()),  # starting error is zero
                0.0,  # starting size is zero
                rng_key_i,
            )

            log_likelihood_i, _, _, _ = eqx.internal.while_loop(
                lambda state: jnp.less_equal(state[1], minimum_mc_error_sq),
                while_body_fn,
                while_body_fn(state_0),  # this makes it a do-while loop
                kind="checkpointed",
                checkpoints=n_checkpoints,
                max_steps=n_max_steps - 1,  # already did one step
            )

            return carry + log_likelihood_i, None

        keys = jrd.split(rng_key, (n_events,))

        total_log_likelihood, _ = jax.lax.scan(
            scan_fn,  # type: ignore[arg-type]
            n_events * jnp.log(T_obs),
            (mean_stack, scale_tril_stack, fit_mean, fit_scale_tril, keys),
            length=n_events,
        )

        # log L(Λ,κ) = -μ + Σ log Σ exp (log ρ(data_n|Λ,κ)) - Σ log(M_i)
        log_likelihood = total_log_likelihood - expected_rates
        # log p(Λ,κ|data) = log π(Λ,κ) + log L(Λ,κ)
        log_posterior = priors.log_prob(x) + log_likelihood

        log_posterior = jnp.nan_to_num(
            log_posterior,
            nan=-jnp.inf,
            posinf=-jnp.inf,
            neginf=-jnp.inf,
        )

        return log_posterior

    return eqx.filter_jit(log_likelihood_fn)



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from . import (
    checks as checks,
    common as common,
    genie_parser as genie_parser,
    ppd as ppd,
    ppd_parser as ppd_parser,
    priors as priors,
    regex as regex,
)



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from typing import Optional

import jax
from jax import lax, numpy as jnp
from jax.scipy.stats import truncnorm
from jaxtyping import Array, ArrayLike
from numpyro.distributions import constraints, Distribution, HalfNormal
from numpyro.distributions.util import promote_shapes, validate_sample

from ...utils.kernel import log_planck_taper_window
from ..constraints import all_constraint, mass_sandwich
from ..redshift import PowerlawRedshift
from ..utils import (
    doubly_truncated_power_law_log_norm_constant,
    JointDistribution,
    ScaledMixture,
)


class BrokenPowerlawTwoPeakMultiSpinMultiTilt(Distribution):
    arg_constraints = {
        "alpha1": constraints.real,
        "alpha2": constraints.real,
        "beta": constraints.real,
        "delta_m1": constraints.positive,
        "delta_m2": constraints.positive,
        "lambda_0": constraints.unit_interval,
        "lambda_1": constraints.unit_interval,
        "loc1": constraints.positive,
        "loc2": constraints.positive,
        "m1min": constraints.real,
        "m2min": constraints.real,
        "mbreak": constraints.real,
        "mmax": constraints.real,
        "scale1": constraints.positive,
        "scale2": constraints.positive,
        "a_1_loc_bpl1": constraints.unit_interval,
        "a_1_loc_bpl2": constraints.unit_interval,
        "a_1_loc_n1": constraints.unit_interval,
        "a_1_loc_n2": constraints.unit_interval,
        "a_2_loc_bpl1": constraints.unit_interval,
        "a_2_loc_bpl2": constraints.unit_interval,
        "a_2_loc_n1": constraints.unit_interval,
        "a_2_loc_n2": constraints.unit_interval,
        "a_1_scale_bpl1": constraints.positive,
        "a_1_scale_bpl2": constraints.positive,
        "a_1_scale_n1": constraints.positive,
        "a_1_scale_n2": constraints.positive,
        "a_2_scale_bpl1": constraints.positive,
        "a_2_scale_bpl2": constraints.positive,
        "a_2_scale_n1": constraints.positive,
        "a_2_scale_n2": constraints.positive,
        "cos_tilt_1_loc_bpl1": constraints.interval(-1.0, 1.0),
        "cos_tilt_1_loc_bpl2": constraints.interval(-1.0, 1.0),
        "cos_tilt_1_loc_n1": constraints.interval(-1.0, 1.0),
        "cos_tilt_1_loc_n2": constraints.interval(-1.0, 1.0),
        "cos_tilt_1_scale_bpl1": constraints.positive,
        "cos_tilt_1_scale_bpl2": constraints.positive,
        "cos_tilt_1_scale_n1": constraints.positive,
        "cos_tilt_1_scale_n2": constraints.positive,
        "cos_tilt_2_loc_bpl1": constraints.interval(-1.0, 1.0),
        "cos_tilt_2_loc_bpl2": constraints.interval(-1.0, 1.0),
        "cos_tilt_2_loc_n1": constraints.interval(-1.0, 1.0),
        "cos_tilt_2_loc_n2": constraints.interval(-1.0, 1.0),
        "cos_tilt_2_scale_bpl1": constraints.positive,
        "cos_tilt_2_scale_bpl2": constraints.positive,
        "cos_tilt_2_scale_n1": constraints.positive,
        "cos_tilt_2_scale_n2": constraints.positive,
        "cos_tilt_zeta_bpl1": constraints.unit_interval,
        "cos_tilt_zeta_bpl2": constraints.unit_interval,
        "cos_tilt_zeta_n1": constraints.unit_interval,
        "cos_tilt_zeta_n2": constraints.unit_interval,
    }

    pytree_data_fields = (
        "_logZ",
        "_m1s",
        "_support",
        "_Z_q_given_m1",
        "alpha1",
        "alpha2",
        "beta",
        "delta_m1",
        "delta_m2",
        "lambda_0",
        "lambda_1",
        "loc1",
        "loc2",
        "m1min",
        "m2min",
        "mbreak",
        "mmax",
        "scale1",
        "scale2",
        "a_1_loc_bpl1",
        "a_1_loc_bpl2",
        "a_1_loc_n1",
        "a_1_loc_n2",
        "a_2_loc_bpl1",
        "a_2_loc_bpl2",
        "a_2_loc_n1",
        "a_2_loc_n2",
        "a_1_scale_bpl1",
        "a_1_scale_bpl2",
        "a_1_scale_n1",
        "a_1_scale_n2",
        "a_2_scale_bpl1",
        "a_2_scale_bpl2",
        "a_2_scale_n1",
        "a_2_scale_n2",
        "cos_tilt_1_loc_bpl1",
        "cos_tilt_1_loc_bpl2",
        "cos_tilt_1_loc_n1",
        "cos_tilt_1_loc_n2",
        "cos_tilt_1_scale_bpl1",
        "cos_tilt_1_scale_bpl2",
        "cos_tilt_1_scale_n1",
        "cos_tilt_1_scale_n2",
        "cos_tilt_2_loc_bpl1",
        "cos_tilt_2_loc_bpl2",
        "cos_tilt_2_loc_n1",
        "cos_tilt_2_loc_n2",
        "cos_tilt_2_scale_bpl1",
        "cos_tilt_2_scale_bpl2",
        "cos_tilt_2_scale_n1",
        "cos_tilt_2_scale_n2",
        "cos_tilt_zeta_bpl1",
        "cos_tilt_zeta_bpl2",
        "cos_tilt_zeta_n1",
        "cos_tilt_zeta_n2",
    )

    def __init__(
        self,
        alpha1: ArrayLike,
        alpha2: ArrayLike,
        beta: ArrayLike,
        loc1: ArrayLike,
        loc2: ArrayLike,
        scale1: ArrayLike,
        scale2: ArrayLike,
        delta_m1: ArrayLike,
        delta_m2: ArrayLike,
        lambda_0: ArrayLike,
        lambda_1: ArrayLike,
        m1min: ArrayLike,
        m2min: ArrayLike,
        mmax: ArrayLike,
        mbreak: ArrayLike,
        a_1_loc_bpl1: ArrayLike,
        a_1_loc_bpl2: ArrayLike,
        a_1_loc_n1: ArrayLike,
        a_1_loc_n2: ArrayLike,
        a_2_loc_bpl1: ArrayLike,
        a_2_loc_bpl2: ArrayLike,
        a_2_loc_n1: ArrayLike,
        a_2_loc_n2: ArrayLike,
        a_1_scale_bpl1: ArrayLike,
        a_1_scale_bpl2: ArrayLike,
        a_1_scale_n1: ArrayLike,
        a_1_scale_n2: ArrayLike,
        a_2_scale_bpl1: ArrayLike,
        a_2_scale_bpl2: ArrayLike,
        a_2_scale_n1: ArrayLike,
        a_2_scale_n2: ArrayLike,
        cos_tilt_1_loc_bpl1: ArrayLike,
        cos_tilt_1_loc_bpl2: ArrayLike,
        cos_tilt_1_loc_n1: ArrayLike,
        cos_tilt_1_loc_n2: ArrayLike,
        cos_tilt_1_scale_bpl1: ArrayLike,
        cos_tilt_1_scale_bpl2: ArrayLike,
        cos_tilt_1_scale_n1: ArrayLike,
        cos_tilt_1_scale_n2: ArrayLike,
        cos_tilt_2_loc_bpl1: ArrayLike,
        cos_tilt_2_loc_bpl2: ArrayLike,
        cos_tilt_2_loc_n1: ArrayLike,
        cos_tilt_2_loc_n2: ArrayLike,
        cos_tilt_2_scale_bpl1: ArrayLike,
        cos_tilt_2_scale_bpl2: ArrayLike,
        cos_tilt_2_scale_n1: ArrayLike,
        cos_tilt_2_scale_n2: ArrayLike,
        cos_tilt_zeta_bpl1: ArrayLike,
        cos_tilt_zeta_bpl2: ArrayLike,
        cos_tilt_zeta_n1: ArrayLike,
        cos_tilt_zeta_n2: ArrayLike,
        validate_args: Optional[bool] = None,
    ) -> None:
        (
            self.alpha1,
            self.alpha2,
            self.beta,
            self.loc1,
            self.loc2,
            self.scale1,
            self.scale2,
            self.delta_m1,
            self.delta_m2,
            self.lambda_0,
            self.lambda_1,
            self.m1min,
            self.m2min,
            self.mmax,
            self.mbreak,
            self.a_1_loc_bpl1,
            self.a_1_loc_bpl2,
            self.a_1_loc_n1,
            self.a_1_loc_n2,
            self.a_2_loc_bpl1,
            self.a_2_loc_bpl2,
            self.a_2_loc_n1,
            self.a_2_loc_n2,
            self.a_1_scale_bpl1,
            self.a_1_scale_bpl2,
            self.a_1_scale_n1,
            self.a_1_scale_n2,
            self.a_2_scale_bpl1,
            self.a_2_scale_bpl2,
            self.a_2_scale_n1,
            self.a_2_scale_n2,
            self.cos_tilt_1_loc_bpl1,
            self.cos_tilt_1_loc_bpl2,
            self.cos_tilt_1_loc_n1,
            self.cos_tilt_1_loc_n2,
            self.cos_tilt_1_scale_bpl1,
            self.cos_tilt_1_scale_bpl2,
            self.cos_tilt_1_scale_n1,
            self.cos_tilt_1_scale_n2,
            self.cos_tilt_2_loc_bpl1,
            self.cos_tilt_2_loc_bpl2,
            self.cos_tilt_2_loc_n1,
            self.cos_tilt_2_loc_n2,
            self.cos_tilt_2_scale_bpl1,
            self.cos_tilt_2_scale_bpl2,
            self.cos_tilt_2_scale_n1,
            self.cos_tilt_2_scale_n2,
            self.cos_tilt_zeta_bpl1,
            self.cos_tilt_zeta_bpl2,
            self.cos_tilt_zeta_n1,
            self.cos_tilt_zeta_n2,
        ) = promote_shapes(
            alpha1,
            alpha2,
            beta,
            loc1,
            loc2,
            scale1,
            scale2,
            delta_m1,
            delta_m2,
            lambda_0,
            lambda_1,
            m1min,
            m2min,
            mmax,
            mbreak,
            a_1_loc_bpl1,
            a_1_loc_bpl2,
            a_1_loc_n1,
            a_1_loc_n2,
            a_2_loc_bpl1,
            a_2_loc_bpl2,
            a_2_loc_n1,
            a_2_loc_n2,
            a_1_scale_bpl1,
            a_1_scale_bpl2,
            a_1_scale_n1,
            a_1_scale_n2,
            a_2_scale_bpl1,
            a_2_scale_bpl2,
            a_2_scale_n1,
            a_2_scale_n2,
            cos_tilt_1_loc_bpl1,
            cos_tilt_1_loc_bpl2,
            cos_tilt_1_loc_n1,
            cos_tilt_1_loc_n2,
            cos_tilt_1_scale_bpl1,
            cos_tilt_1_scale_bpl2,
            cos_tilt_1_scale_n1,
            cos_tilt_1_scale_n2,
            cos_tilt_2_loc_bpl1,
            cos_tilt_2_loc_bpl2,
            cos_tilt_2_loc_n1,
            cos_tilt_2_loc_n2,
            cos_tilt_2_scale_bpl1,
            cos_tilt_2_scale_bpl2,
            cos_tilt_2_scale_n1,
            cos_tilt_2_scale_n2,
            cos_tilt_zeta_bpl1,
            cos_tilt_zeta_bpl2,
            cos_tilt_zeta_n1,
            cos_tilt_zeta_n2,
        )
        batch_shape = lax.broadcast_shapes(
            jnp.shape(alpha1),
            jnp.shape(alpha2),
            jnp.shape(beta),
            jnp.shape(loc1),
            jnp.shape(loc2),
            jnp.shape(scale1),
            jnp.shape(scale2),
            jnp.shape(delta_m1),
            jnp.shape(delta_m2),
            jnp.shape(lambda_0),
            jnp.shape(lambda_1),
            jnp.shape(m1min),
            jnp.shape(m2min),
            jnp.shape(mmax),
            jnp.shape(mbreak),
            jnp.shape(a_1_loc_bpl1),
            jnp.shape(a_1_loc_bpl2),
            jnp.shape(a_1_loc_n1),
            jnp.shape(a_1_loc_n2),
            jnp.shape(a_2_loc_bpl1),
            jnp.shape(a_2_loc_bpl2),
            jnp.shape(a_2_loc_n1),
            jnp.shape(a_2_loc_n2),
            jnp.shape(a_1_scale_bpl1),
            jnp.shape(a_1_scale_bpl2),
            jnp.shape(a_1_scale_n1),
            jnp.shape(a_1_scale_n2),
            jnp.shape(a_2_scale_bpl1),
            jnp.shape(a_2_scale_bpl2),
            jnp.shape(a_2_scale_n1),
            jnp.shape(a_2_scale_n2),
            jnp.shape(cos_tilt_1_loc_bpl1),
            jnp.shape(cos_tilt_1_loc_bpl2),
            jnp.shape(cos_tilt_1_loc_n1),
            jnp.shape(cos_tilt_1_loc_n2),
            jnp.shape(cos_tilt_1_scale_bpl1),
            jnp.shape(cos_tilt_1_scale_bpl2),
            jnp.shape(cos_tilt_1_scale_n1),
            jnp.shape(cos_tilt_1_scale_n2),
            jnp.shape(cos_tilt_2_loc_bpl1),
            jnp.shape(cos_tilt_2_loc_bpl2),
            jnp.shape(cos_tilt_2_loc_n1),
            jnp.shape(cos_tilt_2_loc_n2),
            jnp.shape(cos_tilt_2_scale_bpl1),
            jnp.shape(cos_tilt_2_scale_bpl2),
            jnp.shape(cos_tilt_2_scale_n1),
            jnp.shape(cos_tilt_2_scale_n2),
            jnp.shape(cos_tilt_zeta_bpl1),
            jnp.shape(cos_tilt_zeta_bpl2),
            jnp.shape(cos_tilt_zeta_n1),
            jnp.shape(cos_tilt_zeta_n2),
        )

        self._support = all_constraint(
            [
                mass_sandwich(m2min, mmax),
                constraints.unit_interval,
                constraints.unit_interval,
                constraints.interval(-1.0, 1.0),
                constraints.interval(-1.0, 1.0),
            ],
            [(0, 2), 2, 3, 4, 5],
        )
        super(BrokenPowerlawTwoPeakMultiSpinMultiTilt, self).__init__(
            batch_shape=batch_shape, event_shape=(6,), validate_args=validate_args
        )

        m1min = jnp.broadcast_to(m1min, batch_shape)
        m2min = jnp.broadcast_to(m2min, batch_shape)
        mmax = jnp.broadcast_to(mmax, batch_shape)

        m1s = jnp.linspace(m1min, mmax, 1000)
        self._logZ = jnp.log(
            jnp.trapezoid(
                jnp.exp(self._log_prob_m1_unnorm_component(m1s)).sum(axis=0),
                m1s,
                axis=0,
            )
        )

        self._m1s = jnp.linspace(m2min, mmax, 1000)
        _qs = jnp.linspace(0.005, 1.0, 500)
        _m1_grid, qs_grid = jnp.meshgrid(self._m1s, _qs, indexing="ij")

        _prob_q = jnp.exp(self._log_prob_q_unnorm(_m1_grid, qs_grid))

        self._Z_q_given_m1 = jnp.trapezoid(_prob_q, _qs, axis=1)

    @constraints.dependent_property(is_discrete=False, event_dim=1)
    def support(self) -> constraints.Constraint:
        return self._support

    def _log_prob_m1_unnorm_component(self, m1: Array) -> Array:
        safe_delta = jnp.where(self.delta_m1 <= 0.0, 1.0, self.delta_m1)
        log_smoothing_m1 = log_planck_taper_window((m1 - self.m1min) / safe_delta)
        log_mbreak = jnp.log(self.mbreak)
        log_m1 = jnp.log(m1)
        log_norm_bpl = jnp.logaddexp(
            self.alpha1 * log_mbreak
            + doubly_truncated_power_law_log_norm_constant(
                -self.alpha1, self.m1min, self.mbreak
            ),
            self.alpha2 * log_mbreak
            + doubly_truncated_power_law_log_norm_constant(
                -self.alpha2, self.mbreak, self.mmax
            ),
        )
        log_prob_bpl_1 = jnp.where(
            m1 < self.mbreak, self.alpha1 * (log_mbreak - log_m1), -jnp.inf
        )
        log_prob_bpl_2 = jnp.where(
            m1 < self.mbreak, -jnp.inf, self.alpha2 * (log_mbreak - log_m1)
        )
        log_prob_norm_0 = truncnorm.logpdf(
            m1,
            a=(self.m1min - self.loc1) / self.scale1,
            b=(self.mmax - self.loc1) / self.scale1,
            loc=self.loc1,
            scale=self.scale1,
        )
        log_prob_norm_1 = truncnorm.logpdf(
            m1,
            a=(self.m1min - self.loc2) / self.scale2,
            b=(self.mmax - self.loc2) / self.scale2,
            loc=self.loc2,
            scale=self.scale2,
        )

        log_prob_m1_component = jnp.asarray(
            [
                jnp.log(self.lambda_0)
                + log_prob_bpl_1
                - log_norm_bpl
                + log_smoothing_m1,
                jnp.log(self.lambda_0)
                + log_prob_bpl_2
                - log_norm_bpl
                + log_smoothing_m1,
                jnp.log(self.lambda_1) + log_prob_norm_0 + log_smoothing_m1,
                jnp.log1p(-self.lambda_0 - self.lambda_1)
                + log_prob_norm_1
                + log_smoothing_m1,
            ]
        )

        return jnp.where(
            (self.delta_m1 <= 0.0) | (m1 < self.m1min),
            -jnp.inf,
            log_prob_m1_component,
        )

    def _log_prob_q_unnorm(self, m1: Array, q: Array) -> ArrayLike:
        safe_delta = jnp.where(self.delta_m2 <= 0.0, 1.0, self.delta_m2)
        log_smoothing_q = log_planck_taper_window((m1 * q - self.m2min) / safe_delta)
        log_prob_q = self.beta * jnp.log(q) + log_smoothing_q
        return jnp.where(
            (self.delta_m2 <= 0.0) | (m1 * q < self.m2min), -jnp.inf, log_prob_q
        )

    def _log_prob_a1_components(self, a1: ArrayLike) -> ArrayLike:
        comp_bpl1 = truncnorm.logpdf(
            a1,
            a=(0.0 - self.a_1_loc_bpl1) / self.a_1_scale_bpl1,
            b=(1.0 - self.a_1_loc_bpl1) / self.a_1_scale_bpl1,
            loc=self.a_1_loc_bpl1,
            scale=self.a_1_scale_bpl1,
        )
        comp_bpl2 = truncnorm.logpdf(
            a1,
            a=(0.0 - self.a_1_loc_bpl2) / self.a_1_scale_bpl2,
            b=(1.0 - self.a_1_loc_bpl2) / self.a_1_scale_bpl2,
            loc=self.a_1_loc_bpl2,
            scale=self.a_1_scale_bpl2,
        )
        comp_n1 = truncnorm.logpdf(
            a1,
            a=(0.0 - self.a_1_loc_n1) / self.a_1_scale_n1,
            b=(1.0 - self.a_1_loc_n1) / self.a_1_scale_n1,
            loc=self.a_1_loc_n1,
            scale=self.a_1_scale_n1,
        )
        comp_n2 = truncnorm.logpdf(
            a1,
            a=(0.0 - self.a_1_loc_n2) / self.a_1_scale_n2,
            b=(1.0 - self.a_1_loc_n2) / self.a_1_scale_n2,
            loc=self.a_1_loc_n2,
            scale=self.a_1_scale_n2,
        )
        return jnp.asarray([comp_bpl1, comp_bpl2, comp_n1, comp_n2])

    def _log_prob_a2_components(self, a2: ArrayLike) -> ArrayLike:
        comp_bpl1 = truncnorm.logpdf(
            a2,
            a=(0.0 - self.a_2_loc_bpl1) / self.a_2_scale_bpl1,
            b=(1.0 - self.a_2_loc_bpl1) / self.a_2_scale_bpl1,
            loc=self.a_2_loc_bpl1,
            scale=self.a_2_scale_bpl1,
        )
        comp_bpl2 = truncnorm.logpdf(
            a2,
            a=(0.0 - self.a_2_loc_bpl2) / self.a_2_scale_bpl2,
            b=(1.0 - self.a_2_loc_bpl2) / self.a_2_scale_bpl2,
            loc=self.a_2_loc_bpl2,
            scale=self.a_2_scale_bpl2,
        )
        comp_n1 = truncnorm.logpdf(
            a2,
            a=(0.0 - self.a_2_loc_n1) / self.a_2_scale_n1,
            b=(1.0 - self.a_2_loc_n1) / self.a_2_scale_n1,
            loc=self.a_2_loc_n1,
            scale=self.a_2_scale_n1,
        )
        comp_n2 = truncnorm.logpdf(
            a2,
            a=(0.0 - self.a_2_loc_n2) / self.a_2_scale_n2,
            b=(1.0 - self.a_2_loc_n2) / self.a_2_scale_n2,
            loc=self.a_2_loc_n2,
            scale=self.a_2_scale_n2,
        )
        return jnp.asarray([comp_bpl1, comp_bpl2, comp_n1, comp_n2])

    def _log_prob_t1_t2_components(self, t1: ArrayLike, t2: ArrayLike) -> ArrayLike:
        hyper_params = [
            (
                self.cos_tilt_zeta_bpl1,
                self.cos_tilt_1_loc_bpl1,
                self.cos_tilt_1_scale_bpl1,
                self.cos_tilt_2_loc_bpl1,
                self.cos_tilt_2_scale_bpl1,
            ),
            (
                self.cos_tilt_zeta_bpl2,
                self.cos_tilt_1_loc_bpl2,
                self.cos_tilt_1_scale_bpl2,
                self.cos_tilt_2_loc_bpl2,
                self.cos_tilt_2_scale_bpl2,
            ),
            (
                self.cos_tilt_zeta_n1,
                self.cos_tilt_1_loc_n1,
                self.cos_tilt_1_scale_n1,
                self.cos_tilt_2_loc_n1,
                self.cos_tilt_2_scale_n1,
            ),
            (
                self.cos_tilt_zeta_n2,
                self.cos_tilt_1_loc_n2,
                self.cos_tilt_1_scale_n2,
                self.cos_tilt_2_loc_n2,
                self.cos_tilt_2_scale_n2,
            ),
        ]

        comp_log_probs = []

        for zeta, loc1, scale1, loc2, scale2 in hyper_params:
            comp_gaussian = (
                jnp.log(zeta)
                + truncnorm.logpdf(
                    t1,
                    a=(-1.0 - loc1) / scale1,
                    b=(1.0 - loc1) / scale1,
                    loc=loc1,
                    scale=scale1,
                )
                + truncnorm.logpdf(
                    t2,
                    a=(-1.0 - loc2) / scale2,
                    b=(1.0 - loc2) / scale2,
                    loc=loc2,
                    scale=scale2,
                )
            )
            comp_uniform = jnp.log1p(-zeta) + jnp.log(0.25)
            comp_log_prob = jnp.logaddexp(comp_gaussian, comp_uniform)
            comp_log_probs.append(comp_log_prob)

        return jnp.asarray(comp_log_probs)

    @validate_sample
    def log_prob(self, value: Array) -> ArrayLike:
        m1, m2, a1, a2, t1, t2 = jnp.unstack(value, axis=-1)
        log_prob_m1_component = self._log_prob_m1_unnorm_component(m1)
        log_prob_a1_component = self._log_prob_a1_components(a1)
        log_prob_a2_component = self._log_prob_a2_components(a2)
        log_prob_t1_t2_component = self._log_prob_t1_t2_components(t1, t2)

        log_prob_m1_a1_a2_t1_t2_component = (
            log_prob_m1_component
            + log_prob_a1_component
            + log_prob_a2_component
            + log_prob_t1_t2_component
        )

        log_prob_m1_a1_a2_t1_t2 = jax.nn.logsumexp(
            log_prob_m1_a1_a2_t1_t2_component,
            where=~jnp.isneginf(log_prob_m1_a1_a2_t1_t2_component),
            axis=0,
        )

        _Z_q = jnp.interp(m1, self._m1s, self._Z_q_given_m1, left=1.0, right=1.0)
        safe_Z_q = jnp.where(_Z_q <= 0, 1.0, _Z_q)
        log_Z_q = jnp.where(_Z_q <= 0, 0.0, jnp.log(safe_Z_q))
        log_prob_q = self._log_prob_q_unnorm(m1, m2 / m1) - log_Z_q

        return jnp.log(m1) + log_prob_m1_a1_a2_t1_t2 + log_prob_q - self._logZ


def BrokenPowerlawTwoPeakMultiSpinMultiTiltFull(
    use_eccentricity: bool = False,
    validate_args: Optional[bool] = None,
    **params: Array,
) -> ScaledMixture:
    smoothing_model = BrokenPowerlawTwoPeakMultiSpinMultiTilt(
        alpha1=params["alpha1"],
        alpha2=params["alpha2"],
        beta=params["beta"],
        loc1=params["loc1"],
        loc2=params["loc2"],
        scale1=params["scale1"],
        scale2=params["scale2"],
        delta_m1=params["delta_m1"],
        delta_m2=params["delta_m2"],
        lambda_0=params["lambda_0"],
        lambda_1=params["lambda_1"],
        m1min=params["m1min"],
        m2min=params["m2min"],
        mmax=params["mmax"],
        mbreak=params["mbreak"],
        a_1_loc_bpl1=params["a_1_loc_bpl1"],
        a_1_loc_bpl2=params["a_1_loc_bpl2"],
        a_1_loc_n1=params["a_1_loc_n1"],
        a_1_loc_n2=params["a_1_loc_n2"],
        a_2_loc_bpl1=params["a_2_loc_bpl1"],
        a_2_loc_bpl2=params["a_2_loc_bpl2"],
        a_2_loc_n1=params["a_2_loc_n1"],
        a_2_loc_n2=params["a_2_loc_n2"],
        a_1_scale_bpl1=params["a_1_scale_bpl1"],
        a_1_scale_bpl2=params["a_1_scale_bpl2"],
        a_1_scale_n1=params["a_1_scale_n1"],
        a_1_scale_n2=params["a_1_scale_n2"],
        a_2_scale_bpl1=params["a_2_scale_bpl1"],
        a_2_scale_bpl2=params["a_2_scale_bpl2"],
        a_2_scale_n1=params["a_2_scale_n1"],
        a_2_scale_n2=params["a_2_scale_n2"],
        cos_tilt_1_loc_bpl1=params["cos_tilt_1_loc_bpl1"],
        cos_tilt_1_loc_bpl2=params["cos_tilt_1_loc_bpl2"],
        cos_tilt_1_loc_n1=params["cos_tilt_1_loc_n1"],
        cos_tilt_1_loc_n2=params["cos_tilt_1_loc_n2"],
        cos_tilt_1_scale_bpl1=params["cos_tilt_1_scale_bpl1"],
        cos_tilt_1_scale_bpl2=params["cos_tilt_1_scale_bpl2"],
        cos_tilt_1_scale_n1=params["cos_tilt_1_scale_n1"],
        cos_tilt_1_scale_n2=params["cos_tilt_1_scale_n2"],
        cos_tilt_2_loc_bpl1=params["cos_tilt_2_loc_bpl1"],
        cos_tilt_2_loc_bpl2=params["cos_tilt_2_loc_bpl2"],
        cos_tilt_2_loc_n1=params["cos_tilt_2_loc_n1"],
        cos_tilt_2_loc_n2=params["cos_tilt_2_loc_n2"],
        cos_tilt_2_scale_bpl1=params["cos_tilt_2_scale_bpl1"],
        cos_tilt_2_scale_bpl2=params["cos_tilt_2_scale_bpl2"],
        cos_tilt_2_scale_n1=params["cos_tilt_2_scale_n1"],
        cos_tilt_2_scale_n2=params["cos_tilt_2_scale_n2"],
        cos_tilt_zeta_bpl1=params["cos_tilt_zeta_bpl1"],
        cos_tilt_zeta_bpl2=params["cos_tilt_zeta_bpl2"],
        cos_tilt_zeta_n1=params["cos_tilt_zeta_n1"],
        cos_tilt_zeta_n2=params["cos_tilt_zeta_n2"],
        validate_args=validate_args,
    )

    component_distributions = [smoothing_model]

    if use_eccentricity:
        ecc_dist = HalfNormal(
            scale=params["eccentricity_scale"], validate_args=validate_args
        )
        component_distributions.append(ecc_dist)

    z_max = params["z_max"]
    kappa = params["kappa"]
    powerlaw_z = PowerlawRedshift(z_max=z_max, kappa=kappa, validate_args=validate_args)
    component_distributions.append(powerlaw_z)

    component_distributions = [
        JointDistribution(*component_distributions, validate_args=validate_args)
    ]

    return ScaledMixture(
        log_scales=jnp.asarray([params["log_rate"]]),
        component_distributions=component_distributions,
        support=component_distributions[0]._support,  # type: ignore
        validate_args=validate_args,
    )



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import os
from collections.abc import Callable
from typing import Any, Dict, List

import arviz as az
import jax
import numpy as np
import numpyro
import pandas as pd
from jax import random as jrd
from jaxtyping import Array, PRNGKeyArray
from loguru import logger
from numpyro.infer import MCMC, NUTS

from gwkokab.models.utils import JointDistribution
from gwkokab.utils.tools import warn_if
from kokab.core.guru import Guru, guru_arg_parser
from kokab.utils.common import read_json
from kokab.utils.literals import INFERENCE_DIRECTORY, POSTERIOR_SAMPLES_FILENAME


_INFERENCE_DIRECTORY = "numpyro_" + INFERENCE_DIRECTORY


def _save_inference_data(
    inference_data: az.InferenceData, start_chain_idx: int = 0
) -> None:
    os.makedirs(_INFERENCE_DIRECTORY, exist_ok=True)

    header = list(inference_data.posterior.data_vars.keys())  # type: ignore

    posterior_samples = inference_data.posterior.stack(  # type: ignore
        sample=("chain", "draw")
    ).to_dataframe()

    if start_chain_idx == 0:
        np.savetxt(
            _INFERENCE_DIRECTORY + "/" + POSTERIOR_SAMPLES_FILENAME,
            np.column_stack([posterior_samples[key] for key in header]),
            header=" ".join(header),
        )
    else:
        with open(_INFERENCE_DIRECTORY + "/" + POSTERIOR_SAMPLES_FILENAME, "a") as f:
            np.savetxt(
                f,
                np.column_stack([posterior_samples[key] for key in header]),
            )

    summary = az.summary(inference_data)

    pd.DataFrame(summary).to_json(
        _INFERENCE_DIRECTORY + "/posterior_summary.json", indent=4
    )

    posterior_data = np.permute_dims(
        np.asarray(inference_data.posterior.to_dataarray()),  # type: ignore
        (1, 2, 0),  # (variable, chain, draw) -> (chain, draw, variable)
    )

    n_chains = posterior_data.shape[0]

    for i in range(n_chains):
        np.savetxt(
            _INFERENCE_DIRECTORY + f"/chain_{start_chain_idx + i}.dat",
            posterior_data[i],
            header=" ".join(header),
            comments="#",
            delimiter=" ",
        )


def _run_mcmc(
    key: PRNGKeyArray,
    kernel: numpyro.infer.NUTS,
    mcmc_kwargs: Dict[str, Any],
    data: Dict[str, Any],
):
    n_devices = jax.device_count()
    if (
        chain_method := mcmc_kwargs.pop("chain_method")
    ) != "parallel" and n_devices > 1:
        warn_if(
            True,
            msg=f"Multiple devices detected ({n_devices}), but chain_method is set to "
            f"'{chain_method}'. Overriding to 'parallel'.",
        )
        chain_method = "parallel"

    n_chains = mcmc_kwargs.pop("num_chains", 1)
    batch_size: int = min(n_chains, n_devices)
    n_batches = n_chains // batch_size

    mcmc = MCMC(kernel, num_chains=batch_size, chain_method=chain_method, **mcmc_kwargs)

    chain_idx: int = 0
    for _ in range(n_batches):
        key, subkey = jrd.split(key)
        mcmc.run(subkey, **data)
        inference_data = az.from_numpyro(mcmc)
        _save_inference_data(inference_data, start_chain_idx=chain_idx)
        chain_idx += batch_size

    if n_batches * batch_size < n_chains:
        mcmc.num_chains = n_chains - n_batches * batch_size
        key, subkey = jrd.split(key)
        mcmc.run(subkey, **data)
        inference_data = az.from_numpyro(mcmc)
        _save_inference_data(inference_data, start_chain_idx=chain_idx)


class NumpyroBased(Guru):
    output_directory: str = _INFERENCE_DIRECTORY

    def driver(
        self,
        *,
        logpdf: Callable[[Array, Dict[str, Any]], Array],
        priors: JointDistribution,
        data: Dict[str, Any],
        labels: List[str],
    ) -> None:
        del priors
        del labels

        sampler_config = read_json(self.sampler_settings_filename)

        kernel = NUTS(logpdf, **sampler_config["kernel"])

        if self.debug_nans:
            with jax.debug_nans(True):
                _run_mcmc(self.rng_key, kernel, sampler_config["mcmc"], data)
        elif self.profile_memory:
            _run_mcmc(self.rng_key, kernel, sampler_config["mcmc"], data)

            import datetime

            time = datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
            filename = f"{self.analysis_name}_memory_{time}.prof"
            jax.profiler.save_device_memory_profile(filename)
        elif self.check_leaks:
            with jax.checking_leaks():
                _run_mcmc(self.rng_key, kernel, sampler_config["mcmc"], data)
        else:
            _run_mcmc(self.rng_key, kernel, sampler_config["mcmc"], data)

        logger.info("Sampling and data saving complete.")


numpyro_arg_parser = guru_arg_parser



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from typing import Optional

from jax import numpy as jnp
from jaxtyping import Array
from numpyro.distributions import HalfNormal, Independent

from ..mass import SmoothedTwoComponentPrimaryMassRatio
from ..redshift import PowerlawRedshift
from ..spin import BetaFromMeanVar, IndependentSpinOrientationGaussianIsotropic
from ..transformations import PrimaryMassAndMassRatioToComponentMassesTransform
from ..utils import (
    ExtendedSupportTransformedDistribution,
    JointDistribution,
    ScaledMixture,
)


def PowerlawPeak(
    use_spin: bool = False,
    use_redshift: bool = False,
    use_tilt: bool = False,
    use_eccentricity: bool = False,
    validate_args: Optional[bool] = None,
    **params: Array,
) -> ScaledMixture:
    # NOTE: If you change something here, please also change in
    # kokab/one_powerlaw_one_peak/ppd.py
    smoothing_model = ExtendedSupportTransformedDistribution(
        SmoothedTwoComponentPrimaryMassRatio(
            alpha=params["alpha"],
            beta=params["beta"],
            delta=params["delta"],
            lambda_peak=params["lambda_peak"],
            loc=params["loc"],
            mmax=params["mmax"],
            mmin=params["mmin"],
            scale=params["scale"],
            validate_args=validate_args,
        ),
        transforms=PrimaryMassAndMassRatioToComponentMassesTransform(),
        validate_args=validate_args,
    )

    component_distributions = [smoothing_model]

    if use_spin:
        chi_dist = Independent(
            BetaFromMeanVar(
                mean=jnp.stack([params["chi_mean"], params["chi_mean"]], axis=-1),
                variance=jnp.stack(
                    [params["chi_variance"], params["chi_variance"]], axis=-1
                ),
                validate_args=validate_args,
            ),
            reinterpreted_batch_ndims=1,
            validate_args=validate_args,
        )
        component_distributions.append(chi_dist)

    if use_tilt:
        tilt_dist = IndependentSpinOrientationGaussianIsotropic(
            zeta=params["cos_tilt_zeta"],
            scale1=params["cos_tilt_scale"],
            scale2=params["cos_tilt_scale"],
            validate_args=validate_args,
        )

        component_distributions.append(tilt_dist)

    if use_eccentricity:
        ecc_dist = HalfNormal(
            scale=params["eccentricity_scale"],
            validate_args=validate_args,
        )
        component_distributions.append(ecc_dist)

    if use_redshift:
        z_max = params["z_max"]
        kappa = params["kappa"]
        powerlaw_z = PowerlawRedshift(
            z_max=z_max, kappa=kappa, validate_args=validate_args
        )

        component_distributions.append(powerlaw_z)

    if len(component_distributions) > 1:
        component_distributions = [
            JointDistribution(*component_distributions, validate_args=validate_args)
        ]

    return ScaledMixture(
        log_scales=jnp.asarray([params["log_rate"]]),
        component_distributions=component_distributions,
        support=component_distributions[0]._support,  # type: ignore
        validate_args=validate_args,
    )



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


import json
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from functools import partial
from typing import List, Tuple

from jax import numpy as jnp, random as jrd
from loguru import logger
from numpyro import distributions as dist

from gwkokab.errors import banana_error_m1_m2, truncated_normal_error
from gwkokab.models import NPowerlawMGaussian
from gwkokab.parameters import Parameters as P
from gwkokab.poisson_mean import get_selection_fn_and_poisson_mean_estimator
from kokab.core.population import error_magazine, PopulationFactory
from kokab.utils import genie_parser
from kokab.utils.common import expand_arguments, read_json
from kokab.utils.logger import log_info
from kokab.utils.regex import match_all


def make_parser() -> ArgumentParser:
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser = genie_parser.get_parser(parser)

    model_group = parser.add_argument_group("Model Options")

    model_group.add_argument(
        "--model-json",
        help="Path to the JSON file containing the model parameters",
        type=str,
        required=True,
    )

    spin_magnitude_group = model_group.add_mutually_exclusive_group()
    spin_magnitude_group.add_argument(
        "--add-beta-spin-magnitude",
        action="store_true",
        help="Include beta-distributed spin magnitudes a1,a2 (dimensionless Kerr spins; 0≤a<1).",
    )
    spin_magnitude_group.add_argument(
        "--add-truncated-normal-spin-magnitude",
        action="store_true",
        help="Include truncated-normal spin magnitudes a1,a2 (dimensionless Kerr spins; 0≤a<1).",
    )

    model_group.add_argument(
        "--add-truncated-normal-spin-x",
        action="store_true",
        help="Include truncated-normal spin x components.",
    )
    model_group.add_argument(
        "--add-truncated-normal-spin-y",
        action="store_true",
        help="Include truncated-normal spin y components.",
    )
    model_group.add_argument(
        "--add-truncated-normal-spin-z",
        action="store_true",
        help="Include truncated-normal spin z components.",
    )
    model_group.add_argument(
        "--add-tilt",
        action="store_true",
        help="Include spin-orbit tilt cosines cos_tilt1, cos_tilt2 (cosines of angles between each spin and orbital angular momentum; physical range -1 to 1).",
    )
    model_group.add_argument(
        "--add-truncated-normal-eccentricity",
        action="store_true",
        help="Include orbital eccentricity e via a truncated-normal prior (dimensionless; physical range 0≤e<1 at the reference frequency/time).",
    )
    model_group.add_argument(
        "--add-mean-anomaly",
        action="store_true",
        help="Include mean anomaly M (Kepler mean orbital phase; radians; domain [0,2π)).",
    )
    model_group.add_argument(
        "--add-redshift",
        action="store_true",
        help="Include cosmological redshift z (dimensionless).",
    )
    model_group.add_argument(
        "--add-cos-iota",
        action="store_true",
        help="Include inclination cosine cos(iota) (cosine of angle between orbital angular momentum and line of sight; isotropy ⇒ Uniform[-1,1]).",
    )
    model_group.add_argument(
        "--add-polarization-angle",
        action="store_true",
        help="Include GW polarization angle ψ (radians; domain [0,π); period π).",
    )
    model_group.add_argument(
        "--add-right-ascension",
        action="store_true",
        help="Include right ascension RA (sky longitude; radians; domain [0,2π); uniform for isotropic sky).",
    )
    model_group.add_argument(
        "--add-sin-declination",
        action="store_true",
        help="Include sin(declination) (store sine of declination to ensure uniform sky prior; domain [-1,1]).",
    )
    model_group.add_argument(
        "--add-detection-time",
        action="store_true",
        help="Include geocentric coalescence time (detection_time); typically seconds or GPS time; choose your window in model JSON.",
    )
    model_group.add_argument(
        "--add-phi-1",
        action="store_true",
        help="Include spin azimuth φ₁ (radians; domain [0,2π); affects precession).",
    )
    model_group.add_argument(
        "--add-phi-2",
        action="store_true",
        help="Include spin azimuth φ₂ (radians; domain [0,2π); affects precession).",
    )
    model_group.add_argument(
        "--add-phi-12",
        action="store_true",
        help="Include relative spin azimuth φ₁₂ (radians; domain [0,2π); precession-sensitive).",
    )
    model_group.add_argument(
        "--add-phi-orb",
        action="store_true",
        help="Include orbital phase φ_orb at the reference time/frequency (radians; domain [0,2π)).",
    )

    err_group = parser.add_argument_group("Error Options")
    err_group.add_argument(
        "--err-json",
        help="Path to error JSON defining Gaussian/truncated-normal error scales and bounds (periodic angles are wrapped mod π/2π).",
        type=str,
        required=True,
    )

    return parser


def main() -> None:
    """Main function of the script."""
    parser = make_parser()
    args = parser.parse_args()

    log_info(start=True)

    with open(args.model_json, "r") as f:
        model_json = json.load(f)

    with open(args.err_json, "r") as f:
        err_json = json.load(f)

    N_pl = model_json["N_pl"]
    N_g = model_json["N_g"]

    has_beta_spin_magnitude = args.add_beta_spin_magnitude
    has_truncated_normal_spin_magnitude = args.add_truncated_normal_spin_magnitude
    has_truncated_normal_spin_x = args.add_truncated_normal_spin_x
    has_truncated_normal_spin_y = args.add_truncated_normal_spin_y
    has_truncated_normal_spin_z = args.add_truncated_normal_spin_z
    has_tilt = args.add_tilt
    has_eccentricity = args.add_truncated_normal_eccentricity
    has_mean_anomaly = args.add_mean_anomaly
    has_redshift = args.add_redshift
    has_cos_iota = args.add_cos_iota
    has_polarization_angle = args.add_polarization_angle
    has_right_ascension = args.add_right_ascension
    has_sin_declination = args.add_sin_declination
    has_detection_time = args.add_detection_time
    has_phi_1 = args.add_phi_1
    has_phi_2 = args.add_phi_2
    has_phi_12 = args.add_phi_12
    has_phi_orb = args.add_phi_orb

    err_params_name = ["scale_eta", "scale_Mc"]
    if has_beta_spin_magnitude or has_truncated_normal_spin_magnitude:
        err_params_name.extend(
            [
                P.PRIMARY_SPIN_MAGNITUDE.value + "_high",
                P.PRIMARY_SPIN_MAGNITUDE.value + "_low",
                P.PRIMARY_SPIN_MAGNITUDE.value + "_scale",
                P.SECONDARY_SPIN_MAGNITUDE.value + "_high",
                P.SECONDARY_SPIN_MAGNITUDE.value + "_low",
                P.SECONDARY_SPIN_MAGNITUDE.value + "_scale",
            ]
        )
    if has_truncated_normal_spin_x:
        err_params_name.extend(
            [
                P.PRIMARY_SPIN_X.value + "_high",
                P.PRIMARY_SPIN_X.value + "_low",
                P.PRIMARY_SPIN_X.value + "_scale",
                P.SECONDARY_SPIN_X.value + "_high",
                P.SECONDARY_SPIN_X.value + "_low",
                P.SECONDARY_SPIN_X.value + "_scale",
            ]
        )
    if has_truncated_normal_spin_y:
        err_params_name.extend(
            [
                P.PRIMARY_SPIN_Y.value + "_high",
                P.PRIMARY_SPIN_Y.value + "_low",
                P.PRIMARY_SPIN_Y.value + "_scale",
                P.SECONDARY_SPIN_Y.value + "_high",
                P.SECONDARY_SPIN_Y.value + "_low",
                P.SECONDARY_SPIN_Y.value + "_scale",
            ]
        )
    if has_truncated_normal_spin_z:
        err_params_name.extend(
            [
                P.PRIMARY_SPIN_Z.value + "_high",
                P.PRIMARY_SPIN_Z.value + "_low",
                P.PRIMARY_SPIN_Z.value + "_scale",
                P.SECONDARY_SPIN_Z.value + "_high",
                P.SECONDARY_SPIN_Z.value + "_low",
                P.SECONDARY_SPIN_Z.value + "_scale",
            ]
        )
    if has_tilt:
        err_params_name.extend(
            [
                P.COS_TILT_1.value + "_high",
                P.COS_TILT_1.value + "_low",
                P.COS_TILT_1.value + "_scale",
                P.COS_TILT_2.value + "_high",
                P.COS_TILT_2.value + "_low",
                P.COS_TILT_2.value + "_scale",
            ]
        )
    if has_eccentricity:
        err_params_name.extend(
            [
                P.ECCENTRICITY.value + "_high",
                P.ECCENTRICITY.value + "_low",
                P.ECCENTRICITY.value + "_scale",
            ]
        )
    if has_mean_anomaly:
        err_params_name.extend(
            [
                P.MEAN_ANOMALY.value + "_high",
                P.MEAN_ANOMALY.value + "_low",
                P.MEAN_ANOMALY.value + "_scale",
            ]
        )
    if has_redshift:
        err_params_name.extend(
            [
                P.REDSHIFT.value + "_high",
                P.REDSHIFT.value + "_low",
                P.REDSHIFT.value + "_scale",
            ]
        )
    if has_cos_iota:
        err_params_name.extend(
            [
                P.COS_IOTA.value + "_high",
                P.COS_IOTA.value + "_low",
                P.COS_IOTA.value + "_scale",
            ]
        )
    if has_polarization_angle:
        err_params_name.extend(
            [
                P.POLARIZATION_ANGLE.value + "_high",
                P.POLARIZATION_ANGLE.value + "_low",
                P.POLARIZATION_ANGLE.value + "_scale",
            ]
        )
    if has_right_ascension:
        err_params_name.extend(
            [
                P.RIGHT_ASCENSION.value + "_high",
                P.RIGHT_ASCENSION.value + "_low",
                P.RIGHT_ASCENSION.value + "_scale",
            ]
        )
    if has_sin_declination:
        err_params_name.extend(
            [
                P.SIN_DECLINATION.value + "_high",
                P.SIN_DECLINATION.value + "_low",
                P.SIN_DECLINATION.value + "_scale",
            ]
        )
    if has_detection_time:
        err_params_name.append(P.DETECTION_TIME.value + "_scale")
    if has_phi_1:
        err_params_name.extend(
            [
                P.PHI_1.value + "_high",
                P.PHI_1.value + "_low",
                P.PHI_1.value + "_scale",
            ]
        )
    if has_phi_2:
        err_params_name.extend(
            [
                P.PHI_2.value + "_high",
                P.PHI_2.value + "_low",
                P.PHI_2.value + "_scale",
            ]
        )
    if has_phi_12:
        err_params_name.extend(
            [
                P.PHI_12.value + "_high",
                P.PHI_12.value + "_low",
                P.PHI_12.value + "_scale",
            ]
        )
    if has_phi_orb:
        err_params_name.extend(
            [
                P.PHI_ORB.value + "_high",
                P.PHI_ORB.value + "_low",
                P.PHI_ORB.value + "_scale",
            ]
        )

    err_params_value = match_all(err_params_name, err_json)

    all_params: List[Tuple[str, int]] = [
        ("alpha_pl", N_pl),
        ("beta_pl", N_pl),
        ("log_rate", N_pl + N_g),
        ("m1_high_g", N_g),
        ("m1_loc_g", N_g),
        ("m1_low_g", N_g),
        ("m1_scale_g", N_g),
        ("m2_high_g", N_g),
        ("m2_loc_g", N_g),
        ("m2_low_g", N_g),
        ("m2_scale_g", N_g),
        ("mmax_pl", N_pl),
        ("mmin_pl", N_pl),
    ]

    parameters_name: Tuple[str, ...] = (
        P.PRIMARY_MASS_SOURCE.value,
        P.SECONDARY_MASS_SOURCE.value,
    )

    error_magazine.register(
        (P.PRIMARY_MASS_SOURCE.value, P.SECONDARY_MASS_SOURCE.value),
        lambda x, size, key: banana_error_m1_m2(
            x,
            size,
            key,
            scale_Mc=err_params_value["scale_Mc"],
            scale_eta=err_params_value["scale_eta"],
        ),
    )

    if has_beta_spin_magnitude:
        parameters_name += (
            P.PRIMARY_SPIN_MAGNITUDE.value,
            P.SECONDARY_SPIN_MAGNITUDE.value,
        )
        all_params.extend(
            [
                (P.PRIMARY_SPIN_MAGNITUDE.value + "_mean_g", N_g),
                (P.PRIMARY_SPIN_MAGNITUDE.value + "_mean_pl", N_pl),
                (P.PRIMARY_SPIN_MAGNITUDE.value + "_variance_g", N_g),
                (P.PRIMARY_SPIN_MAGNITUDE.value + "_variance_pl", N_pl),
                (P.SECONDARY_SPIN_MAGNITUDE.value + "_mean_g", N_g),
                (P.SECONDARY_SPIN_MAGNITUDE.value + "_mean_pl", N_pl),
                (P.SECONDARY_SPIN_MAGNITUDE.value + "_variance_g", N_g),
                (P.SECONDARY_SPIN_MAGNITUDE.value + "_variance_pl", N_pl),
            ]
        )

    if has_truncated_normal_spin_magnitude:
        parameters_name += (
            P.PRIMARY_SPIN_MAGNITUDE.value,
            P.SECONDARY_SPIN_MAGNITUDE.value,
        )
        all_params.extend(
            [
                (P.PRIMARY_SPIN_MAGNITUDE.value + "_high_g", N_g),
                (P.PRIMARY_SPIN_MAGNITUDE.value + "_high_pl", N_pl),
                (P.PRIMARY_SPIN_MAGNITUDE.value + "_loc_g", N_g),
                (P.PRIMARY_SPIN_MAGNITUDE.value + "_loc_pl", N_pl),
                (P.PRIMARY_SPIN_MAGNITUDE.value + "_low_g", N_g),
                (P.PRIMARY_SPIN_MAGNITUDE.value + "_low_pl", N_pl),
                (P.PRIMARY_SPIN_MAGNITUDE.value + "_scale_g", N_g),
                (P.PRIMARY_SPIN_MAGNITUDE.value + "_scale_pl", N_pl),
                (P.SECONDARY_SPIN_MAGNITUDE.value + "_high_g", N_g),
                (P.SECONDARY_SPIN_MAGNITUDE.value + "_high_pl", N_pl),
                (P.SECONDARY_SPIN_MAGNITUDE.value + "_loc_g", N_g),
                (P.SECONDARY_SPIN_MAGNITUDE.value + "_loc_pl", N_pl),
                (P.SECONDARY_SPIN_MAGNITUDE.value + "_low_g", N_g),
                (P.SECONDARY_SPIN_MAGNITUDE.value + "_low_pl", N_pl),
                (P.SECONDARY_SPIN_MAGNITUDE.value + "_scale_g", N_g),
                (P.SECONDARY_SPIN_MAGNITUDE.value + "_scale_pl", N_pl),
            ]
        )

    if has_beta_spin_magnitude or has_truncated_normal_spin_magnitude:
        error_magazine.register(
            P.PRIMARY_SPIN_MAGNITUDE.value,
            partial(
                truncated_normal_error,
                scale=err_params_value[P.PRIMARY_SPIN_MAGNITUDE.value + "_scale"],
                low=err_params_value.get(P.PRIMARY_SPIN_MAGNITUDE.value + "_low"),
                high=err_params_value.get(P.PRIMARY_SPIN_MAGNITUDE.value + "_high"),
                cut_low=0.0,
                cut_high=1.0,
            ),
        )

        error_magazine.register(
            P.SECONDARY_SPIN_MAGNITUDE.value,
            partial(
                truncated_normal_error,
                scale=err_params_value[P.SECONDARY_SPIN_MAGNITUDE.value + "_scale"],
                low=err_params_value.get(P.SECONDARY_SPIN_MAGNITUDE.value + "_low"),
                high=err_params_value.get(P.SECONDARY_SPIN_MAGNITUDE.value + "_high"),
                cut_low=0.0,
                cut_high=1.0,
            ),
        )

    if has_truncated_normal_spin_x:
        parameters_name += (P.PRIMARY_SPIN_X.value, P.SECONDARY_SPIN_X.value)
        all_params.extend(
            [
                (P.PRIMARY_SPIN_X.value + "_high_g", N_g),
                (P.PRIMARY_SPIN_X.value + "_high_pl", N_pl),
                (P.PRIMARY_SPIN_X.value + "_loc_g", N_g),
                (P.PRIMARY_SPIN_X.value + "_loc_pl", N_pl),
                (P.PRIMARY_SPIN_X.value + "_low_g", N_g),
                (P.PRIMARY_SPIN_X.value + "_low_pl", N_pl),
                (P.PRIMARY_SPIN_X.value + "_scale_g", N_g),
                (P.PRIMARY_SPIN_X.value + "_scale_pl", N_pl),
                (P.SECONDARY_SPIN_X.value + "_high_g", N_g),
                (P.SECONDARY_SPIN_X.value + "_high_pl", N_pl),
                (P.SECONDARY_SPIN_X.value + "_loc_g", N_g),
                (P.SECONDARY_SPIN_X.value + "_loc_pl", N_pl),
                (P.SECONDARY_SPIN_X.value + "_low_g", N_g),
                (P.SECONDARY_SPIN_X.value + "_low_pl", N_pl),
                (P.SECONDARY_SPIN_X.value + "_scale_g", N_g),
                (P.SECONDARY_SPIN_X.value + "_scale_pl", N_pl),
            ]
        )

        error_magazine.register(
            P.PRIMARY_SPIN_X.value,
            partial(
                truncated_normal_error,
                scale=err_params_value[P.PRIMARY_SPIN_X.value + "_scale"],
                low=err_params_value.get(P.PRIMARY_SPIN_X.value + "_low"),
                high=err_params_value.get(P.PRIMARY_SPIN_X.value + "_high"),
                cut_low=-1.0,
                cut_high=1.0,
            ),
        )

        error_magazine.register(
            P.SECONDARY_SPIN_X.value,
            partial(
                truncated_normal_error,
                scale=err_params_value[P.SECONDARY_SPIN_X.value + "_scale"],
                low=err_params_value.get(P.SECONDARY_SPIN_X.value + "_low"),
                high=err_params_value.get(P.SECONDARY_SPIN_X.value + "_high"),
                cut_low=-1.0,
                cut_high=1.0,
            ),
        )

    if has_truncated_normal_spin_y:
        parameters_name += (P.PRIMARY_SPIN_Y.value, P.SECONDARY_SPIN_Y.value)
        all_params.extend(
            [
                (P.PRIMARY_SPIN_Y.value + "_high_g", N_g),
                (P.PRIMARY_SPIN_Y.value + "_high_pl", N_pl),
                (P.PRIMARY_SPIN_Y.value + "_loc_g", N_g),
                (P.PRIMARY_SPIN_Y.value + "_loc_pl", N_pl),
                (P.PRIMARY_SPIN_Y.value + "_low_g", N_g),
                (P.PRIMARY_SPIN_Y.value + "_low_pl", N_pl),
                (P.PRIMARY_SPIN_Y.value + "_scale_g", N_g),
                (P.PRIMARY_SPIN_Y.value + "_scale_pl", N_pl),
                (P.SECONDARY_SPIN_Y.value + "_high_g", N_g),
                (P.SECONDARY_SPIN_Y.value + "_high_pl", N_pl),
                (P.SECONDARY_SPIN_Y.value + "_loc_g", N_g),
                (P.SECONDARY_SPIN_Y.value + "_loc_pl", N_pl),
                (P.SECONDARY_SPIN_Y.value + "_low_g", N_g),
                (P.SECONDARY_SPIN_Y.value + "_low_pl", N_pl),
                (P.SECONDARY_SPIN_Y.value + "_scale_g", N_g),
                (P.SECONDARY_SPIN_Y.value + "_scale_pl", N_pl),
            ]
        )

        error_magazine.register(
            P.PRIMARY_SPIN_Y.value,
            partial(
                truncated_normal_error,
                scale=err_params_value[P.PRIMARY_SPIN_Y.value + "_scale"],
                low=err_params_value.get(P.PRIMARY_SPIN_Y.value + "_low"),
                high=err_params_value.get(P.PRIMARY_SPIN_Y.value + "_high"),
                cut_low=-1.0,
                cut_high=1.0,
            ),
        )

        error_magazine.register(
            P.SECONDARY_SPIN_Y.value,
            partial(
                truncated_normal_error,
                scale=err_params_value[P.SECONDARY_SPIN_Y.value + "_scale"],
                low=err_params_value.get(P.SECONDARY_SPIN_Y.value + "_low"),
                high=err_params_value.get(P.SECONDARY_SPIN_Y.value + "_high"),
                cut_low=-1.0,
                cut_high=1.0,
            ),
        )

    if has_truncated_normal_spin_z:
        parameters_name += (P.PRIMARY_SPIN_Z.value, P.SECONDARY_SPIN_Z.value)
        all_params.extend(
            [
                (P.PRIMARY_SPIN_Z.value + "_high_g", N_g),
                (P.PRIMARY_SPIN_Z.value + "_high_pl", N_pl),
                (P.PRIMARY_SPIN_Z.value + "_loc_g", N_g),
                (P.PRIMARY_SPIN_Z.value + "_loc_pl", N_pl),
                (P.PRIMARY_SPIN_Z.value + "_low_g", N_g),
                (P.PRIMARY_SPIN_Z.value + "_low_pl", N_pl),
                (P.PRIMARY_SPIN_Z.value + "_scale_g", N_g),
                (P.PRIMARY_SPIN_Z.value + "_scale_pl", N_pl),
                (P.SECONDARY_SPIN_Z.value + "_high_g", N_g),
                (P.SECONDARY_SPIN_Z.value + "_high_pl", N_pl),
                (P.SECONDARY_SPIN_Z.value + "_loc_g", N_g),
                (P.SECONDARY_SPIN_Z.value + "_loc_pl", N_pl),
                (P.SECONDARY_SPIN_Z.value + "_low_g", N_g),
                (P.SECONDARY_SPIN_Z.value + "_low_pl", N_pl),
                (P.SECONDARY_SPIN_Z.value + "_scale_g", N_g),
                (P.SECONDARY_SPIN_Z.value + "_scale_pl", N_pl),
            ]
        )

        error_magazine.register(
            P.PRIMARY_SPIN_Z.value,
            partial(
                truncated_normal_error,
                scale=err_params_value[P.PRIMARY_SPIN_Z.value + "_scale"],
                low=err_params_value.get(P.PRIMARY_SPIN_Z.value + "_low"),
                high=err_params_value.get(P.PRIMARY_SPIN_Z.value + "_high"),
                cut_low=-1.0,
                cut_high=1.0,
            ),
        )

        error_magazine.register(
            P.SECONDARY_SPIN_Z.value,
            partial(
                truncated_normal_error,
                scale=err_params_value[P.SECONDARY_SPIN_Z.value + "_scale"],
                low=err_params_value.get(P.SECONDARY_SPIN_Z.value + "_low"),
                high=err_params_value.get(P.SECONDARY_SPIN_Z.value + "_high"),
                cut_low=-1.0,
                cut_high=1.0,
            ),
        )

    if has_tilt:
        parameters_name += (P.COS_TILT_1.value, P.COS_TILT_2.value)
        all_params.extend(
            [
                ("cos_tilt_zeta_g", N_g),
                ("cos_tilt_zeta_pl", N_pl),
                (P.COS_TILT_1.value + "_scale_g", N_g),
                (P.COS_TILT_1.value + "_scale_pl", N_pl),
                (P.COS_TILT_2.value + "_scale_g", N_g),
                (P.COS_TILT_2.value + "_scale_pl", N_pl),
            ]
        )

        error_magazine.register(
            P.COS_TILT_1.value,
            partial(
                truncated_normal_error,
                scale=err_params_value[P.COS_TILT_1.value + "_scale"],
                low=err_params_value.get(P.COS_TILT_1.value + "_low"),
                high=err_params_value.get(P.COS_TILT_1.value + "_high"),
                cut_low=-1.0,
                cut_high=1.0,
            ),
        )

        error_magazine.register(
            P.COS_TILT_2.value,
            partial(
                truncated_normal_error,
                scale=err_params_value[P.COS_TILT_2.value + "_scale"],
                low=err_params_value.get(P.COS_TILT_2.value + "_low"),
                high=err_params_value.get(P.COS_TILT_2.value + "_high"),
                cut_low=-1.0,
                cut_high=1.0,
            ),
        )

    if has_phi_1:
        parameters_name += (P.PHI_1.value,)

        all_params.extend(
            [
                (P.PHI_1.value + "_high_g", N_g),
                (P.PHI_1.value + "_high_pl", N_pl),
                (P.PHI_1.value + "_low_g", N_g),
                (P.PHI_1.value + "_low_pl", N_pl),
            ]
        )

        def phi_1_error(x, size, key):
            err_x = dist.TruncatedNormal(
                loc=x,
                scale=err_params_value[P.PHI_1.value + "_scale"],
                low=err_params_value.get(P.PHI_1.value + "_low", 0.0),
                high=err_params_value.get(P.PHI_1.value + "_high", 2 * jnp.pi),
            ).sample(key=key, sample_shape=(size,))

            err_x = jnp.mod(err_x, 2 * jnp.pi)
            return err_x

        error_magazine.register(P.PHI_1.value, phi_1_error)

    if has_phi_2:
        parameters_name += (P.PHI_2.value,)

        all_params.extend(
            [
                (P.PHI_2.value + "_high_g", N_g),
                (P.PHI_2.value + "_high_pl", N_pl),
                (P.PHI_2.value + "_low_g", N_g),
                (P.PHI_2.value + "_low_pl", N_pl),
            ]
        )

        def phi_2_error(x, size, key):
            err_x = dist.TruncatedNormal(
                loc=x,
                scale=err_params_value[P.PHI_2.value + "_scale"],
                low=err_params_value.get(P.PHI_2.value + "_low", 0.0),
                high=err_params_value.get(P.PHI_2.value + "_high", 2 * jnp.pi),
            ).sample(key=key, sample_shape=(size,))

            err_x = jnp.mod(err_x, 2 * jnp.pi)
            return err_x

        error_magazine.register(P.PHI_2.value, phi_2_error)

    if has_phi_12:
        parameters_name += (P.PHI_12.value,)

        all_params.extend(
            [
                (P.PHI_12.value + "_high_g", N_g),
                (P.PHI_12.value + "_high_pl", N_pl),
                (P.PHI_12.value + "_low_g", N_g),
                (P.PHI_12.value + "_low_pl", N_pl),
            ]
        )

        def phi_12_error(x, size, key):
            err_x = dist.TruncatedNormal(
                loc=x,
                scale=err_params_value[P.PHI_12.value + "_scale"],
                low=err_params_value.get(P.PHI_12.value + "_low", 0.0),
                high=err_params_value.get(P.PHI_12.value + "_high", 2 * jnp.pi),
            ).sample(key=key, sample_shape=(size,))

            err_x = jnp.mod(err_x, 2 * jnp.pi)
            return err_x

        error_magazine.register(P.PHI_12.value, phi_12_error)

    if has_eccentricity:
        parameters_name += (P.ECCENTRICITY.value,)
        all_params.extend(
            [
                (P.ECCENTRICITY.value + "_high_g", N_g),
                (P.ECCENTRICITY.value + "_high_pl", N_pl),
                (P.ECCENTRICITY.value + "_loc_g", N_g),
                (P.ECCENTRICITY.value + "_loc_pl", N_pl),
                (P.ECCENTRICITY.value + "_low_g", N_g),
                (P.ECCENTRICITY.value + "_low_pl", N_pl),
                (P.ECCENTRICITY.value + "_scale_g", N_g),
                (P.ECCENTRICITY.value + "_scale_pl", N_pl),
            ]
        )

        error_magazine.register(
            P.ECCENTRICITY.value,
            partial(
                truncated_normal_error,
                scale=err_params_value[P.ECCENTRICITY.value + "_scale"],
                low=err_params_value.get(P.ECCENTRICITY.value + "_low"),
                high=err_params_value.get(P.ECCENTRICITY.value + "_high"),
                cut_low=0.0,
                cut_high=1.0,
            ),
        )

    if has_mean_anomaly:
        parameters_name += (P.MEAN_ANOMALY.value,)

        all_params.extend(
            [
                (P.MEAN_ANOMALY.value + "_high_g", N_g),
                (P.MEAN_ANOMALY.value + "_high_pl", N_pl),
                (P.MEAN_ANOMALY.value + "_low_g", N_g),
                (P.MEAN_ANOMALY.value + "_low_pl", N_pl),
            ]
        )

        def mean_anomaly_error(x, size, key):
            err_x = dist.TruncatedNormal(
                loc=x,
                scale=err_params_value[P.MEAN_ANOMALY.value + "_scale"],
                low=err_params_value.get(P.MEAN_ANOMALY.value + "_low", 0.0),
                high=err_params_value.get(P.MEAN_ANOMALY.value + "_high", 2.0 * jnp.pi),
            ).sample(key=key, sample_shape=(size,))
            err_x = jnp.mod(err_x, 2.0 * jnp.pi)
            return err_x

        error_magazine.register(P.MEAN_ANOMALY.value, mean_anomaly_error)

    if has_redshift:
        parameters_name += (P.REDSHIFT.value,)

        all_params.extend(
            [
                (P.REDSHIFT.value + "_kappa_g", N_g),
                (P.REDSHIFT.value + "_kappa_pl", N_pl),
                (P.REDSHIFT.value + "_z_max_g", N_g),
                (P.REDSHIFT.value + "_z_max_pl", N_pl),
            ]
        )

        error_magazine.register(
            P.REDSHIFT.value,
            partial(
                truncated_normal_error,
                scale=err_params_value[P.REDSHIFT.value + "_scale"],
                low=err_params_value.get(P.REDSHIFT.value + "_low"),
                high=err_params_value.get(P.REDSHIFT.value + "_high"),
                cut_low=1e-3,
                cut_high=None,
            ),
        )

    if has_right_ascension:
        parameters_name += (P.RIGHT_ASCENSION.value,)

        all_params.extend(
            [
                (P.RIGHT_ASCENSION.value + "_high_g", N_g),
                (P.RIGHT_ASCENSION.value + "_high_pl", N_pl),
                (P.RIGHT_ASCENSION.value + "_low_g", N_g),
                (P.RIGHT_ASCENSION.value + "_low_pl", N_pl),
            ]
        )

        def right_ascension_error(x, size, key):
            err_x = dist.TruncatedNormal(
                loc=x,
                scale=err_params_value[P.RIGHT_ASCENSION.value + "_scale"],
                low=err_params_value.get(P.RIGHT_ASCENSION.value + "_low", 0.0),
                high=err_params_value.get(
                    P.RIGHT_ASCENSION.value + "_high", 2.0 * jnp.pi
                ),
            ).sample(key=key, sample_shape=(size,))
            err_x = jnp.mod(err_x, 2.0 * jnp.pi)
            return err_x

        error_magazine.register(P.RIGHT_ASCENSION.value, right_ascension_error)

    if has_sin_declination:
        parameters_name += (P.SIN_DECLINATION.value,)

        all_params.extend(
            [
                (P.SIN_DECLINATION.value + "_high_g", N_g),
                (P.SIN_DECLINATION.value + "_high_pl", N_pl),
                (P.SIN_DECLINATION.value + "_low_g", N_g),
                (P.SIN_DECLINATION.value + "_low_pl", N_pl),
            ]
        )

        error_magazine.register(
            P.SIN_DECLINATION.value,
            partial(
                truncated_normal_error,
                scale=err_params_value[P.SIN_DECLINATION.value + "_scale"],
                low=err_params_value.get(P.SIN_DECLINATION.value + "_low"),
                high=err_params_value.get(P.SIN_DECLINATION.value + "_high"),
                cut_low=-1.0,
                cut_high=1.0,
            ),
        )

    if has_detection_time:
        parameters_name += (P.DETECTION_TIME.value,)

        all_params.extend(
            [
                (P.DETECTION_TIME.value + "_high_g", N_g),
                (P.DETECTION_TIME.value + "_high_pl", N_pl),
                (P.DETECTION_TIME.value + "_low_g", N_g),
                (P.DETECTION_TIME.value + "_low_pl", N_pl),
            ]
        )

        def detection_time_error(x, size, key):
            eps = 1e-6  # To avoid log(0) or log of negative
            safe_x = jnp.maximum(x, eps)
            err_x = dist.LogNormal(
                loc=jnp.log(safe_x),
                scale=err_params_value[P.DETECTION_TIME.value + "_scale"],
            ).sample(key=key, sample_shape=(size,))

            return err_x

        error_magazine.register(P.DETECTION_TIME.value, detection_time_error)

    if has_cos_iota:
        parameters_name += (P.COS_IOTA.value,)

        all_params.extend(
            [
                (P.COS_IOTA.value + "_high_g", N_g),
                (P.COS_IOTA.value + "_high_pl", N_pl),
                (P.COS_IOTA.value + "_low_g", N_g),
                (P.COS_IOTA.value + "_low_pl", N_pl),
            ]
        )

        error_magazine.register(
            P.COS_IOTA.value,
            partial(
                truncated_normal_error,
                scale=err_params_value[P.COS_IOTA.value + "_scale"],
                low=err_params_value.get(P.COS_IOTA.value + "_low"),
                high=err_params_value.get(P.COS_IOTA.value + "_high"),
                cut_low=-1.0,
                cut_high=1.0,
            ),
        )

    if has_polarization_angle:
        parameters_name += (P.POLARIZATION_ANGLE.value,)

        all_params.extend(
            [
                (P.POLARIZATION_ANGLE.value + "_high_g", N_g),
                (P.POLARIZATION_ANGLE.value + "_high_pl", N_pl),
                (P.POLARIZATION_ANGLE.value + "_low_g", N_g),
                (P.POLARIZATION_ANGLE.value + "_low_pl", N_pl),
            ]
        )

        def polarization_angle_error(x, size, key):
            err_x = dist.TruncatedNormal(
                loc=x,
                scale=err_params_value[P.POLARIZATION_ANGLE.value + "_scale"],
                low=err_params_value.get(P.POLARIZATION_ANGLE.value + "_low", 0.0),
                high=err_params_value.get(P.POLARIZATION_ANGLE.value + "_high", jnp.pi),
            ).sample(key=key, sample_shape=(size,))
            err_x = jnp.mod(err_x, jnp.pi)
            return err_x

        error_magazine.register(P.POLARIZATION_ANGLE.value, polarization_angle_error)

    if has_phi_orb:
        parameters_name += (P.PHI_ORB.value,)

        all_params.extend(
            [
                (P.PHI_ORB.value + "_high_g", N_g),
                (P.PHI_ORB.value + "_high_pl", N_pl),
                (P.PHI_ORB.value + "_low_g", N_g),
                (P.PHI_ORB.value + "_low_pl", N_pl),
            ]
        )

        def phi_orb_error(x, size, key):
            err_x = dist.TruncatedNormal(
                loc=x,
                scale=err_params_value[P.PHI_ORB.value + "_scale"],
                low=err_params_value.get(P.PHI_ORB.value + "_low", 0.0),
                high=err_params_value.get(P.PHI_ORB.value + "_high", 2 * jnp.pi),
            ).sample(key=key, sample_shape=(size,))

            err_x = jnp.mod(err_x, 2 * jnp.pi)
            return err_x

        error_magazine.register(P.PHI_ORB.value, phi_orb_error)

    extended_params = []
    for params in all_params:
        extended_params.extend(expand_arguments(*params))

    model_param = match_all(extended_params, model_json)

    model_param.update(
        {
            "N_pl": N_pl,
            "N_g": N_g,
            "use_beta_spin_magnitude": has_beta_spin_magnitude,
            "use_truncated_normal_spin_magnitude": has_truncated_normal_spin_magnitude,
            "use_truncated_normal_spin_x": has_truncated_normal_spin_x,
            "use_truncated_normal_spin_y": has_truncated_normal_spin_y,
            "use_truncated_normal_spin_z": has_truncated_normal_spin_z,
            "use_tilt": has_tilt,
            "use_eccentricity": has_eccentricity,
            "use_mean_anomaly": has_mean_anomaly,
            "use_redshift": has_redshift,
            "use_cos_iota": has_cos_iota,
            "use_phi_12": has_phi_12,
            "use_polarization_angle": has_polarization_angle,
            "use_right_ascension": has_right_ascension,
            "use_sin_declination": has_sin_declination,
            "use_detection_time": has_detection_time,
            "use_phi_1": has_phi_1,
            "use_phi_2": has_phi_2,
            "use_phi_orb": has_phi_orb,
        }
    )

    logger.info(f"Setting the random number generator key with seed {args.seed}.")
    pmean_key, factory_key = jrd.split(jrd.PRNGKey(args.seed), 2)

    pmean_config = read_json(args.pmean_json)
    log_selection_fn, erate_estimator, _, _ = (
        get_selection_fn_and_poisson_mean_estimator(
            key=pmean_key, parameters=parameters_name, **pmean_config
        )
    )

    popfactory = PopulationFactory(
        model_fn=NPowerlawMGaussian,
        model_params=model_param,
        parameters=parameters_name,
        log_selection_fn=log_selection_fn,
        poisson_mean_estimator=erate_estimator,
        num_realizations=args.num_realizations,
        error_size=args.error_size,
        tile_covariance=args.tile_covariance,
    )
    popfactory.produce(factory_key)



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from collections.abc import Callable
from typing import Dict, List, Optional, Tuple

import equinox as eqx
import jax
import numpyro
from jax import Array, numpy as jnp
from jaxtyping import ArrayLike
from numpyro._typing import DistributionT
from numpyro.distributions import Distribution

from ..models.utils import JointDistribution, LazyJointDistribution, ScaledMixture


__all__ = ["numpyro_poisson_likelihood"]


def numpyro_poisson_likelihood(
    dist_fn: Callable[..., DistributionT],
    priors: JointDistribution,
    variables: Dict[str, DistributionT],
    variables_index: Dict[str, int],
    log_constants: ArrayLike,
    poisson_mean_estimator: Callable[[ScaledMixture], Array],
    where_fns: Optional[List[Callable[..., Array]]],
    constants: Dict[str, Array],
) -> Callable[[List[Array], List[Array], List[Array]], Array]:
    is_lazy_prior = isinstance(priors, LazyJointDistribution)
    if is_lazy_prior:
        dependencies = priors.dependencies
        partial_order = priors.partial_order
    del priors

    def log_likelihood_fn(
        data_group: Tuple[Array, ...],
        log_ref_priors_group: Tuple[Array, ...],
        masks_group: Tuple[Array, ...],
    ):
        if is_lazy_prior:
            partial_variables_samples = [
                numpyro.sample(parameter_name, prior_dist)
                if isinstance(prior_dist, Distribution)
                else (parameter_name, prior_dist)
                for parameter_name, prior_dist in sorted(
                    variables.items(), key=lambda x: x[0]
                )
            ]
            for i in partial_order:
                kwargs = {
                    k: partial_variables_samples[v] for k, v in dependencies[i].items()
                }
                parameter_name, prior_dist_fn = partial_variables_samples[i]
                if isinstance(prior_dist_fn, jax.tree_util.Partial):
                    prior_dist = prior_dist_fn.func(
                        *prior_dist_fn.args, **prior_dist_fn.keywords, **kwargs
                    )  # type: ignore
                partial_variables_samples[i] = numpyro.sample(
                    parameter_name, prior_dist
                )
            variables_samples = partial_variables_samples  # type: ignore
        else:
            variables_samples = [
                numpyro.sample(parameter_name, prior_dist)
                for parameter_name, prior_dist in sorted(
                    variables.items(), key=lambda x: x[0]
                )
            ]
        mapped_params = {
            name: variables_samples[i] for name, i in variables_index.items()
        }

        model_instance: DistributionT = dist_fn(**mapped_params, validate_args=True)

        # μ = E_{θ|Λ}[VT(θ)]
        expected_rates = eqx.filter_jit(poisson_mean_estimator)(model_instance)

        n_buckets = len(masks_group)

        @jax.jit
        def _total_log_likelihood_fn(*args: Array):
            data_group = args[0:n_buckets]
            log_ref_priors_group = args[n_buckets : 2 * n_buckets]
            masks_group = args[2 * n_buckets : 3 * n_buckets]

            total_log_likelihood = log_constants  # - Σ log(M_i)
            # Σ log Σ exp (log p(θ|data_n) - log π_n)
            for batched_data, batched_log_ref_priors, batched_masks in zip(
                data_group, log_ref_priors_group, masks_group
            ):
                safe_data = jnp.where(
                    jnp.expand_dims(batched_masks, axis=-1),
                    batched_data,
                    model_instance.support.feasible_like(batched_data),
                )
                safe_log_ref_prior = jnp.where(
                    batched_masks, batched_log_ref_priors, 0.0
                )

                n_events_per_bucket, n_samples, _ = batched_data.shape
                batched_model_log_prob = jax.vmap(
                    jax.vmap(model_instance.log_prob, axis_size=n_samples),
                    axis_size=n_events_per_bucket,
                )(safe_data)  # type: ignore
                safe_model_log_prob = jnp.where(
                    batched_masks, batched_model_log_prob, -jnp.inf
                )
                batched_log_prob: Array = safe_model_log_prob - safe_log_ref_prior
                batched_log_prob = jnp.where(
                    batched_masks & (~jnp.isnan(batched_log_prob)),
                    batched_log_prob,
                    -jnp.inf,
                )
                log_prob_sum = jax.nn.logsumexp(
                    batched_log_prob,
                    axis=-1,
                    where=~jnp.isneginf(batched_log_prob),
                )
                safe_log_prob_sum = jnp.where(
                    jnp.isneginf(log_prob_sum), -jnp.inf, log_prob_sum
                )
                total_log_likelihood += jnp.sum(safe_log_prob_sum, axis=-1)

            if where_fns is not None and len(where_fns) > 0:
                mask = where_fns[0](**constants, **mapped_params)
                for where_fn in where_fns[1:]:
                    mask = mask & where_fn(**constants, **mapped_params)
                total_log_likelihood = jnp.where(
                    mask,
                    total_log_likelihood,
                    -jnp.inf,  # type: ignore
                )
            return total_log_likelihood

        # - μ + Σ log Σ exp (log p(θ|data_n) - log π_n) - Σ log(M_i)
        numpyro.factor(
            "log_likelihood",
            _total_log_likelihood_fn(
                *data_group,
                *log_ref_priors_group,
                *masks_group,
            )
            - expected_rates,
        )

    return log_likelihood_fn  # type: ignore



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from abc import abstractmethod
from argparse import ArgumentParser
from collections import defaultdict
from collections.abc import Callable
from typing import Any, Dict, List, Optional, Set, Tuple, Union

import jax
from jax import lax, random as jrd
from jaxtyping import Array, PRNGKeyArray
from loguru import logger
from numpyro._typing import DistributionT
from numpyro.distributions.distribution import Distribution
from numpyro.util import is_prng_key

from gwkokab.models.utils import JointDistribution, LazyJointDistribution
from gwkokab.utils.tools import error_if
from kokab.utils.common import read_json, write_json
from kokab.utils.priors import get_processed_priors


def _topological_sort(graph: Dict[str, Set[str]]) -> List[str]:
    visited = set()
    result = []

    def dfs(node: str) -> None:
        visited.add(node)
        for neighbor in graph.get(node, set()):
            if neighbor not in visited:
                dfs(neighbor)
        result.append(node)

    for node in graph:
        if node not in visited:
            dfs(node)
    result.reverse()
    return result


def _check_cycles(graph: Dict[str, Set[str]]) -> bool:
    """Checks if a directed graph has cycles using Depth-First Search (DFS).

    Parameters
    ----------
    graph : Dict[str, Set[str]]
        A directed graph represented as an adjacency list.

    Returns
    -------
    bool
        True if the graph has cycles, False otherwise.
    """
    visited = set()
    rec_stack = set()

    def dfs(node: str) -> bool:
        visited.add(node)
        rec_stack.add(node)

        for neighbor in graph.get(node, set()):
            if neighbor not in visited:
                if dfs(neighbor):
                    return True
            elif neighbor in rec_stack:
                return True

        rec_stack.remove(node)
        return False

    for node in graph:
        if node not in visited:
            if dfs(node):
                return True
    return False


def _bake_model(
    dist_factory: Distribution | Callable[..., Distribution], **params: Any
) -> Tuple[
    Dict[str, int | float | None],
    Dict[str, Distribution],
    Dict[str, str],
    Callable[..., Distribution],
    Dict[str, Dict[str, str]],
    List[str],
]:
    """Separate constants, random variables, and lazy variables from model parameters.

    Parameters
    ----------
    dist_factory : Distribution | Callable[..., Distribution]
        A distribution class or a callable that constructs a distribution.

    Returns
    -------
    Tuple[
        Dict[str, int | float | None],
        Dict[str, Distribution],
        Dict[str, str],
        Callable[..., Distribution],
        Dict[str, Dict[str, str]],
        List[str],
    ]
        A tuple containing:
        - constants: fixed numeric or None values
        - variables: distribution instances or lazy callables
        - aliases: mapping of parameter aliases to their original names
        - baked_factory: callable with constants pre-applied
        - lazy_dependencies: mapping of lazy variable definitions
        - lazy_order: topological order of dependent lazy variables

    Raises
    ------
    ValueError
        If a parameter has an invalid type.
    """

    constants: Dict[str, int | float | None] = {}
    variables: Dict[str, Distribution] = {}
    aliases: Dict[str, str] = {}
    lazy_dependencies: Dict[str, Dict[str, str]] = {}

    dependency_graph: Dict[str, Set[str]] = defaultdict(set)
    variable_roots: List[str] = []

    # Pass 1: classify parameters
    for name, value in params.items():
        if value is None:
            constants[name] = None

        elif isinstance(value, Distribution):
            variables[name] = value
            variable_roots.append(name)

        elif isinstance(value, tuple):
            lazy_fn, lazy_args = value
            error_if(
                not isinstance(lazy_fn, jax.tree_util.Partial),
                msg=f"Lazy distribution '{name}' must be a `jax.tree_util.Partial`.",
            )
            error_if(
                not isinstance(lazy_args, dict),
                msg=f"Lazy distribution '{name}' must have a dictionary of dependencies.",
            )

            variables[name] = lazy_fn
            lazy_dependencies[name] = lazy_args

            for _, dep_name in lazy_args.items():
                dependency_graph[dep_name].add(name)

        elif isinstance(value, (int, float)):
            constants[name] = lax.stop_gradient(value)

        elif isinstance(value, str):
            # string aliases handled later
            continue

        else:
            error_if(
                True,
                msg=f"Invalid parameter '{name}' with type {type(value)} and value {value}",
            )

    # Pass 2: resolve string aliases
    for name, value in params.items():
        if isinstance(value, str):
            if value in constants:
                constants[name] = constants[value]
            elif value in variables:
                aliases[name] = value

    # Compute lazy evaluation order
    if lazy_dependencies:
        import pprint

        error_if(
            _check_cycles(dependency_graph),
            msg="Cyclic dependencies detected among lazy variables. Dependency graph:\n"
            + pprint.pformat(dependency_graph),
        )

        lazy_order = [
            var
            for var in _topological_sort(dependency_graph)
            if var not in variable_roots
        ]
    else:
        lazy_order = []

    baked_factory = jax.tree_util.Partial(dist_factory, **constants)

    return (
        constants,
        variables,
        aliases,
        baked_factory,
        lazy_dependencies,
        lazy_order,
    )


class Guru:
    """Guru is a class which contains all the common functionality among Genie, Sage and
    Guru classes.
    """

    _rng_key: PRNGKeyArray
    output_directory: str

    def __init__(
        self,
        *,
        analysis_name: str,
        check_leaks: bool,
        debug_nans: bool,
        model: Union[DistributionT, Callable[..., DistributionT]],
        poisson_mean_filename: str,
        prior_filename: str,
        profile_memory: bool,
        sampler_settings_filename: str,
        variance_cut_threshold: Optional[float] = None,
    ) -> None:
        self.analysis_name = analysis_name
        self.prior_filename = prior_filename
        self.model = model
        self.sampler_settings_filename = sampler_settings_filename
        self.debug_nans = debug_nans
        self.profile_memory = profile_memory
        self.check_leaks = check_leaks
        self.poisson_mean_filename = poisson_mean_filename
        self.variance_cut_threshold = variance_cut_threshold

    @property
    def rng_key(self) -> PRNGKeyArray:
        self._rng_key, subkey = jrd.split(self._rng_key)
        return subkey

    def set_rng_key(
        self, *, key: Optional[PRNGKeyArray] = None, seed: Optional[int] = None
    ) -> None:
        error_if(
            key is None and seed is None,
            msg="Either 'key' or 'seed' must be provided to set the random number generator key.",
        )
        if key is not None:
            error_if(
                not is_prng_key(key),
                msg=f"Expected a PRNGKeyArray, got {type(key)}.",
            )
            logger.info(f"Setting the random number generator key to {key}.")
            self._rng_key = key
        elif seed is not None:
            error_if(
                not isinstance(seed, int),
                msg=f"Expected an integer seed, got {type(seed)}.",
            )
            error_if(
                seed < 0,
                msg=f"Seed must be a non-negative integer, got {seed}.",
            )
            logger.info(f"Setting the random number generator key with seed {seed}.")
            key = jrd.PRNGKey(seed)
            self._rng_key = key

    def bake_model(
        self,
    ) -> Tuple[
        Dict[str, Union[int, float, bool, None]],
        Callable[..., DistributionT],
        JointDistribution,
        Dict[str, int],
        Dict[str, int],
    ]:
        """Returns a Bake object for the model.

        Returns
        -------
        Tuple[ Dict[str, Union[int, float, bool, None]], Callable[..., DistributionT], JointDistribution, Dict[str, int], ]
            A tuple containing the constants, the distribution function, the prior
            distribution, and the variables index.
        """
        prior_dict = read_json(self.prior_filename)
        model_prior_param = get_processed_priors(self.model_parameters, prior_dict)

        logger.debug("Baking the model")
        constants, variables, duplicates, dist_fn, lazy_deps, lazy_order = _bake_model(
            self.model, **self.constants, **model_prior_param
        )  # type: ignore

        variables_index: dict[str, int] = {
            key: i for i, key in enumerate(sorted(variables.keys()))
        }
        for key, value in duplicates.items():
            variables_index[key] = variables_index[value]

        group_variables: dict[int, list[str]] = {}
        for key, value in variables_index.items():  # type: ignore
            group_variables[value] = group_variables.get(value, []) + [key]  # type: ignore

        logger.debug(
            "Number of recovering variables: {num_vars}", num_vars=len(group_variables)
        )

        for key, value in constants.items():  # type: ignore
            logger.debug(
                "Constant variable: {name} = {variable}", name=key, variable=value
            )

        for value in group_variables.values():  # type: ignore
            logger.debug("Recovering variable: {variable}", variable=", ".join(value))

        write_json("constants.json", constants)
        write_json("nf_samples_mapping.json", variables_index)

        sorted_variables = sorted(variables.keys())
        if len(lazy_order) == 0:
            priors = JointDistribution(
                *[variables[key] for key in sorted_variables],
                validate_args=True,
            )
        else:
            lazy_deps_new = {
                sorted_variables.index(k): {
                    k_: sorted_variables.index(v) for k_, v in deps.items()
                }
                for k, deps in lazy_deps.items()
            }
            priors = LazyJointDistribution(
                *[variables[key] for key in sorted_variables],
                dependencies=lazy_deps_new,
                partial_order=[sorted_variables.index(k) for k in lazy_order],
                validate_args=True,
            )

        return constants, dist_fn, priors, variables, variables_index

    @property
    def parameters(self) -> List[str]:
        """Returns the parameters (intrinsic + extrinsic).

        Returns
        -------
        List[str]
            List of parameters.

        Raises
        ------
        NotImplementedError
            If the Guru class is used directly, this method raises a NotImplementedError.
            It is expected that subclasses of Guru will implement this method.
        """
        msg = (
            "The Guru class should not be used directly. Please use a subclass that "
            "implements the parameters property."
        )
        logger.error(msg)
        raise NotImplementedError(msg)

    @property
    def model_parameters(self) -> List[str]:
        """Returns the model parameters.

        Returns
        -------
        List[str]
            List of model parameters.

        Raises
        ------
        NotImplementedError
            If the Guru class is used directly, this method raises a NotImplementedError.
            It is expected that subclasses of Guru will implement this method.
        """
        msg = (
            "The Guru class should not be used directly. Please use a subclass that "
            "implements the model_parameters property."
        )
        logger.error(msg)
        raise NotImplementedError(msg)

    @property
    def constants(self) -> Dict[str, Union[int, float, bool]]:
        """Returns the constants used in the model.

        Returns
        -------
        Dict[str, Union[int, float, bool]]
            A dictionary containing the constants used in the model.
        """
        return {}

    @abstractmethod
    def driver(
        self,
        *,
        logpdf: Callable[[Array, Dict[str, Any]], Array],
        priors: JointDistribution,
        data: Any,
        labels: List[str],
    ) -> None:
        raise NotImplementedError()


def guru_arg_parser(parser: ArgumentParser) -> ArgumentParser:
    """Populate the command line argument parser with the arguments for the Guru script.

    Parameters
    ----------
    parser : ArgumentParser
        Parser to add the arguments to

    Returns
    -------
    ArgumentParser
        the command line argument parser
    """

    parser.add_argument(
        "--seed",
        help="Seed for the random number generator.",
        default=37,
        type=int,
    )

    pmean_group = parser.add_argument_group("Poisson Mean Options")
    pmean_group.add_argument(
        "--pmean-json",
        help="Path to the JSON file containing the Poisson mean options.",
        type=str,
        default="pmean.json",
    )

    sampler_group = parser.add_argument_group("Sampler Options")
    sampler_group.add_argument(
        "--sampler-config",
        help="Path to the JSON file containing the sampler configuration.",
        type=str,
        required=True,
    )
    sampler_group.add_argument(
        "--variance-cut-threshold",
        help="Threshold for variance cut in the sampler.",
        type=float,
        default=None,
    )

    prior_group = parser.add_argument_group("Prior Options")
    prior_group.add_argument(
        "--prior-json",
        type=str,
        help="Path to a JSON file containing the prior distributions.",
        default="prior.json",
    )

    dev_group = parser.add_argument_group(
        "Developer Options",
        description="These options are intended for developers. They will not work in "
        "production mode.",
    )
    dev_group.add_argument(
        "--debug-nans",
        help="Checks for NaNs in each computation. See details in the documentation: "
        "https://jax.readthedocs.io/en/latest/_autosummary/jax.debug_nans.html#jax.debug_nans.",
        action="store_true",
    )
    dev_group.add_argument(
        "--profile-memory",
        help="Enable memory profiling.",
        action="store_true",
    )
    dev_group.add_argument(
        "--check-leaks",
        help="Check for JAX Tracer leaks. See details in the documentation: "
        "https://jax.readthedocs.io/en/latest/_autosummary/jax.checking_leaks.html#jax.checking_leaks.",
        action="store_true",
    )

    return parser



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


from ._bp2pfull import BrokenPowerlawTwoPeakFull as BrokenPowerlawTwoPeakFull
from ._bp2pmultispinfull import (
    BrokenPowerlawTwoPeakMultiSpinMultiTilt as BrokenPowerlawTwoPeakMultiSpinMultiTilt,
    BrokenPowerlawTwoPeakMultiSpinMultiTiltFull as BrokenPowerlawTwoPeakMultiSpinMultiTiltFull,
)
from ._npowerlawmgaussian import NPowerlawMGaussian as NPowerlawMGaussian
from ._powerlawpeak import PowerlawPeak as PowerlawPeak



# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


#
"""Core modules of Kokab."""

from . import (
    flowMC_based as flowMC_based,
    guru as guru,
    monk as monk,
    numpyro_based as numpyro_based,
    population as population,
    sage as sage,
)
